import urllib.parse
import json
import io
import time
import signal
import os, sys
import re
import math
import tempfile
import socket
import subprocess
import argparse
import shlex
import atexit
import logging
from collections import defaultdict
from functools import partialmethod, cache
from threading import Thread
from queue import Queue

from .utils import randstr, dictify, listify
from .timestamps import timestamp_parser
from .cli import ThrowingArgumentParserError, ThrowingArgumentParser

import requests
from requests.auth import HTTPBasicAuth

logger = logging.getLogger(__name__)
serve_logger = logging.getLogger(f"{__name__}-rc-server")


class RcloneError(ValueError):
    pass


class RC:
    """
    rclone RC interface.

    Still a bit of a work in progress, especially around documentation.

    Note that all filepaths (e.g. src, dst) will use rcpathsplit which accepts
    a regular path and heuristically breaks it into filesystem (fs) and remote/file or
    specify a 2-tuple of (fs,remote/file). If fs is known as separate, this will be
    optimal as rclone caches information by fs.

    rclone_exe ['rclone']
        Executable to call

    serve_flags [None]
        Flags to always include. Examples includes --config, -vv, etc

        (note if using -vv, set serve_log_callback to something)

    rclone_env (None / Empty dict)
        Environment to awlays include. This is appended to os.environ.
        If there is something in os.environ that is *not* wanted, specify it
        with RLCONE.DELENV. ex:
            {"RCLONE_PASSWORD_COMMAND": RC.DELENV}

    Note:
    -----
    The server outputs to a different logger than the rest so it can be filtered.
    This doesn't make much difference if not started with -vv

    """

    DELENV = "**DELENV**"  # Remove from environment
    NOFLAG = "**NOFLAG**"  # Remove from call (Not sure this is used)

    def __init__(
        self,
        rclone_exe="rclone",
        serve_flags=None,
        rclone_env=None,
        # serve_log_callback=None,
    ):
        self.rclone_exe = rclone_exe
        self.addr = f"localhost:{random_port()}"
        self.serve_flags = listify(serve_flags)
        self.rclone_env = dictify(rclone_env)
        # self.serve_log_callback = serve_log_callback

        self.user = randstr()
        self.password = randstr()

        self._started = False
        self._exit = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.stop()

    def check(self):
        try:
            self.call("rc/noop")
            return True
        except requests.ConnectionError:
            return False

    def start(self, check=False):
        if check:
            self._started = self.check()

        if self._started:
            return self

        logger.debug("Starting rclone rc server")
        logger.debug(f"http://{self.user}:{self.password}@{self.addr}")

        cmd = [self.rclone_exe, "rcd"] + self.serve_flags
        cmd.append("--rc-serve")  # For reading remote content
        cmd.extend(["--rc-addr", self.addr])
        cmd.extend(["--rc-user", self.user])
        cmd.extend(["--rc-pass", self.password])
        cmd.extend(["--rc-server-read-timeout", "100h"])
        cmd.extend(["--rc-server-write-timeout", "100h"])
        cmd.extend(["--log-format", ""])  # dates can be captured with internal logging

        env = os.environ.copy()
        env.update({str(k): str(v) for k, v in self.rclone_env.items()})
        for key in list(env):
            if env[key] == self.DELENV:
                del env[key]
        e = {k: v for k, v in env.items() if k not in os.environ}
        logger.debug(f"rclone call {str(cmd)} with env: {json.dumps(e)}")

        self.proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )

        self._started = True

        self.server_reader_thread = Thread(target=self.server_reader, daemon=True)
        self.server_reader_thread.start()

        self._wait_for_start()
        atexit.register(self.stop)
        return self

    def server_reader(self):
        for oe, line in popen_streamer(self.proc, allow_error=True):
            line = line.decode().rstrip("\n")
            for ll in line.split("\n"):
                # serve_logger.debug(f"{oe}: {ll}")
                serve_logger.debug(ll)

            if self._exit:
                break

    def _wait_for_start(self, dt=0.2, timeout=5):
        n = math.ceil(timeout / dt)
        for i in range(n):
            try:
                self.call("rc/noop")
                logger.debug(f"Started at {i = }")
                break
            except requests.ConnectionError:  # ConnectionError
                pass
            time.sleep(dt)
        else:
            raise ValueError("Failed to start server")

    def stop(self):
        logger.debug("stopping rclone rc server")
        if not self._started:
            return
        self._exit = True

        try:
            self.call("core/quit")
        except:
            pass

        self.proc.send_signal(signal.SIGINT)
        try:
            self.proc.wait(timeout=0.25)
        except:
            try:
                self.proc.send_signal(signal.SIGKILL)
            except:
                pass
        return self

    def _cpmvfile(self, *, cpmv, src, dst, use_async=False, **params):
        """
        src and dst can either be strings or (Fs,Remote) tuples. The latter will
        optimize the calls since it will cache more information
        """
        endpoint = f"operations/{cpmv}"
        method = self.call_async_and_wait if use_async else self.call

        srcFs, srcRemote = rcpathsplit(src)
        dstFs, dstRemote = rcpathsplit(dst)
        params["srcFs"] = srcFs
        params["srcRemote"] = srcRemote
        params["dstFs"] = dstFs
        params["dstRemote"] = dstRemote
        return method(endpoint, params=params)

    copyfile = partialmethod(_cpmvfile, cpmv="copyfile")
    movefile = partialmethod(_cpmvfile, cpmv="movefile")

    def delete(self, file, use_async=False, **params):
        method = self.call_async_and_wait if use_async else self.call

        endpoint = "operations/deletefile"
        params["fs"], params["remote"] = rcpathsplit(file)

        return method(endpoint, params=params)

    def write(self, dst, content, use_async=False, **params):
        endpoint = "operations/uploadfile"
        method = self.call_async_and_wait if use_async else self.call

        params["fs"], name = rcpathsplit(dst)
        params["remote"], name = os.path.split(name)

        if isinstance(content, str):
            content = content.encode()
        try:
            return self.call(
                endpoint,
                params=params,
                postkw=dict(files={name: content}),
            )
        except:
            logger.debug("Trying write fallback")
            params.pop("fs", None)
            params.pop("remote", None)
            return self._write_fallback(dst, content, use_async=use_async, **params)

    def _write_fallback(self, dst, content, use_async=False, **params):
        _config = params.get("_config", {})
        _config["NoCheckDest"] = True
        params["_config"] = _config

        if isinstance(content, str):
            content = content.encode()

        with tempfile.NamedTemporaryFile() as fp:
            fp.write(content)
            fp.flush()

            return self.copyfile(
                src=fp.name,
                dst=dst,
                use_async=use_async,
                **params,
            )

    def _http_head(self, src):
        self.start()

        fs, file = rcpathsplit(src)
        file_url = urllib.parse.urljoin(
            f"http://{self.user}:{self.password}@{self.addr}",
            f"[{fs}]/{file}",
        )
        res = requests.head(file_url)
        return res.headers

    def read(self, src, start=0, end=None):
        """
        Read directly from the remote. This is like "rclone cat"

        Inputs:
        -------

        src
            Either a remote path or (fs,remote/file) tuple

        start [0]
            Starting range

        end [None]
            End range

        Note on start,end
        ----------------
        Ranges are passed directly to as headers and are INCLUSIVE. (i.e. end=255
        is the first 256 bytes). Specify as None to be excluded. For example,

            | start | end  | comment                    |
            |-------|------|----------------------------|
            | 0     | None | Read all                   |
            | 256   | None | Read all but the first 256 |
            | None  | 300  | read the *last* 300 bytes  |

        See https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Range

        If there is a failure, just download the file and read it. But need to be
        careful of ranges
        """
        self.start()

        fs, file = rcpathsplit(src)
        file_url = urllib.parse.urljoin(
            f"http://{self.user}:{self.password}@{self.addr}",
            f"[{fs}]/{file}",
        )

        if start is None:
            start = ""
        if end is None:
            end = ""

        res = requests.get(file_url, headers={"Range": f"bytes={start}-{end}"})
        if res.status_code == 404:
            raise ValueError("Not Found or range too far")
        return res.content

    def open(
        self,
        remotefile,
        mode="rb",
        buffer_size=8 * 1024 * 1024,
    ):
        """
        Return a buffered, seekable, READ-ONLY, file-like object of 'remotefile',
        that is buffered with 'buff_size'. Default buffer_size is 8 MiB or
        8*1024*1024 = 8388608 bytes

        mode:
            Only supports reading so 'r','rt','rb'

        buffer_size [8388608]
            Amount to buffer. Default: 8 MiB = 8*1024*1024 bytes = 8388608 bytes

        """
        if "w" in mode or "a" in mode:
            raise io.UnsupportedOperation("Cannot open in write or append mode")
        raw = _RawRcloneFileObj(remotefile, self)
        fp = io.BufferedReader(raw, buffer_size=buffer_size)
        if "b" not in mode:
            fp = io.TextIOWrapper(fp)

        for attr in ["remotefile", "fs", "remote"]:
            if val := getattr(raw, attr, None):
                setattr(fp, attr, val)
        fp._is_rc_file = True  # used in rcpathsplit

        return fp

    def list(
        self,
        remote,
        *,
        recurse=True,
        filters=None,
        filter_params=None,
        filter_cli="",
        mimetype=False,
        modtime=True,
        hashes=False,
        hashtypes=False,
        metadata=False,
        maxdepth=None,
        only=None,
        epoch_time=False,
        fast_list=True,
        config_params=None,
        use_async=False,
        **params,
    ):
        """
        List remotes. Note that most filters are via filter_params or can be converted
        from the CLI flags with filter_cli. The only exception is the 'filters' keyword
        which gets appended to filter_params['FilterRule']

        Inputs:
        -------
        remote
            Either (Fs,Remote) tuple or a a single string which will be split

        recurse [True]
            Whether to recursively list

        filters:
            Convenience for filter_params['FilterRule']. Everything else is via
            filter_params! List or string

        filter_params [empty]
            Filters for _filter in the rc call. An incomplete list includes:

                    'FilterRule' 'FilterFrom' 'FilesFrom' 'FilesFromRaw'
                    'ExcludeRule' 'ExcludeFrom' 'IncludeRule' 'IncludeFrom' 'ExcludeFile'
                    'MinAge' 'MaxAge' 'MinSize' 'MaxSize'
                    'DeleteExcluded' 'IgnoreCase'

            See 'options/get' for full list and see filter_cli keyword to accept command
            line flags.

            References: https://rclone.org/rc/#setting-filter-flags-with-filter

        filter_cli [None]
            Convenience for filter_params = filter_cli2params(filter_cli). Can be
            a list of a string

        mimetype [False]
            If False, will call with equiv of --no-mimetype

        modtime [True]
            If False, will call with equiv of --no-modtime. This can be faster on remotes
            that require additional requests for it. Note that it still may or may not
            include a ModTime. ModTimes are converted to datetime objects unless
            epoch_time=True.

        hashes [False]
            Compute hashes

        hashtypes [None]
            If specified and hashes=True, will use these types. String of one or list
            of types

        metadata [False]
            If True, equivalent of --metadata

        maxdepth [None]
            Adds equiv of --max-depth flags.

        only [None]
            Specify as {None [default],'files','dirs'}.

        epoch_time [False]
            If True, returns a Unix Epoch float. Otherwise, returns a timezone aware
            datetime object

            These can be set here and/or in the

        fast_list [True]
            Whether or not to use --fast-list. If being recursive, this is probably
            better

        config_params [empty]
           Additional _config dictionary. See options/get

        **params
            Passed to the call(). Use config_params to set stat params
        """
        # Note: Everthing is build into the params keyword argument.
        # So we pop from it as needed for other settings

        # Build the call. First params
        params["fs"], params["remote"] = rcpathsplit(remote)
        params["remote"] = params["remote"].removesuffix("/")

        # Filters. Allow for _filter in **params, FilterRule, filter_params,
        # and filter_cli (in that order) with extra care taken for MetaRules
        filter_params = params.get("_filter", {}) | dictify(filter_params)

        if f := listify(filters):
            filter_params["FilterRule"] = filter_params.get("FilterRule", []) + f

        meta_filters = filter_params.pop("MetaRules", {})

        cli = filter_cli2params(filter_cli)
        meta_filters |= cli.pop("MetaRules", {})

        filter_params |= cli
        if meta_filters:
            filter_params["MetaRules"] = meta_filters

        params["_filter"] = filter_params

        # Other config
        params["_config"] = params.get("_config", {}) | dictify(config_params)

        # Rest of the KWs
        opt = params["opt"] = params.get("opt", {})
        opt["recurse"] = recurse
        opt["noMimeType"] = not mimetype
        opt["noModTime"] = not modtime
        opt["showHash"] = hashes
        if hashtypes and hashes:
            opt["hashTypes"] = listify(hashtypes)
        opt["metadata"] = metadata

        if only is None:
            pass
        elif only == "dirs":
            opt["dirsOnly"] = True
        elif only == "files":
            opt["filesOnly"] = True
        else:
            raise ValueError(f"Inalid {only = }")

        if maxdepth:
            params["_config"]["MaxDepth"] = maxdepth

        params["_config"]["UseListR"] = fast_list

        method = self.call_async_and_wait if use_async else self.call

        res = method("operations/list", params=params)
        if "list" not in res:
            raise ValueError(f"Error listing remote. Check keywords. {res = }")
        for line in res["list"]:
            # Never understood why rclone gives us this...
            line.pop("Name", None)

            if "ModTime" in line:  # Do regardless of modtime setting
                line["ModTime"] = timestamp_parser(line["ModTime"], epoch=epoch_time)
        return res["list"]

    def stat(
        self,
        remotefile,
        *,
        mimetype=False,
        modtime=True,
        hashes=False,
        hashtypes=None,
        metadata=False,
        epoch_time=False,
        config_params=None,
        use_async=False,
        **params,
    ):
        """
        remotefile
            Either a string path to the file or (fs,file) tuple

        mimetype [False]
            If False, will call with equiv of --no-mimetype

        modtime [True]
            If False, will call with equiv of --no-modtime. This can be faster on remotes
            that require additional requests for it. Note that it still may or may not
            include a ModTime. ModTimes are converted to datetime objects unless
            epoch_time=True.

        hashes [False]
            Compute hashes

        hashtypes [None]
            If specified and hashes=True, will use these types. String of one or list
            of types

        metadata [False]
            If True, equivalent of --metadata

        epoch_time [False]
            If True, returns a Unix Epoch float. Otherwise, returns a timezone aware
            datetime object

        config_params [empty]
           Additional _config dictionary. See options/get

        **params
            Passed to the call(). Use config_params to set stat params
        """
        params["fs"], params["remote"] = rcpathsplit(remotefile)

        params["_config"] = params.get("_config", {}) | dictify(config_params)
        params["opt"] = opt = params.get("opt", {})

        opt["filesOnly"] = True
        opt["noMimeType"] = not mimetype
        opt["noModTime"] = not modtime
        opt["showHash"] = hashes
        if hashtypes and hashes:
            opt["hashTypes"] = listify(hashtypes)
        opt["metadata"] = metadata

        method = self.call_async_and_wait if use_async else self.call

        item = self.call("operations/stat", params=params).get("item", None)
        if not item:
            return

        item.pop("Name", None)
        if "ModTime" in item:  # Do regardless of modtime setting
            item["ModTime"] = timestamp_parser(item["ModTime"], epoch=epoch_time)

        return item

    @cache
    def features(self, fs, **params):
        params["fs"] = fs
        return self.call("operations/fsinfo", params=params)

    def call(self, endpoint, *, postkw=None, params=None, **paramskwargs):
        self.start()

        postkw = postkw or {}
        params = paramskwargs | (params or {})

        logging.debug(f"call: {endpoint = }, {params = }")

        for key, val in params.items():
            if isinstance(val, (dict, list)):
                params[key] = json.dumps(val)

        # In order to get sending data for rcat (aka write) to work, we use the URL
        # paramaters and post anything else as data. This makes the URLs more cumbersome
        # but in my testing, works better since you can post content.
        url = (
            urllib.parse.urljoin(f"http://{self.addr}", endpoint)
            + "?"
            + urllib.parse.urlencode(params)
        )

        resp = requests.post(
            url,
            auth=HTTPBasicAuth(self.user, self.password),
            **postkw,
        )
        res = resp.json()

        # This is developer-level debug. Comment out for now
        # logger.debug(f"call {res = }")

        if res.get("error", ""):
            err = RcloneError(f"Error. Result: {res}")
            err.response = res
            raise err
        return res

    def call_async_and_background(
        self, endpoint, postkw=None, params=None, **paramskwargs
    ):
        """Call with async and return jobid"""
        params = paramskwargs | (params or {})
        params = params.copy()
        params["_async"] = True
        jobid = self.call(endpoint, postkw=postkw, params=params)["jobid"]
        return jobid

    def check_async(self, jobid):
        """Check on async job. return none if not done"""
        res = self.call("job/status", params={"jobid": jobid})
        if res.get("finished", False) or res.get("error", ""):
            return res

    def call_async_and_wait(self, endpoint, postkw=None, params=None, **paramskwargs):
        """
        This is basically the same as call() but uses async. Calls and waits
        for return
        """
        jobid = self.call_async_and_background(
            endpoint, postkw=postkw, params=params, **paramskwargs
        )
        t0 = time.time()

        while True:
            elapsed = time.time() - t0
            dt = dtfun(elapsed)
            if res := self.check_async(jobid):
                break
            time.sleep(dt)
        return res


class _RawRcloneFileObj(io.RawIOBase):
    # PRIVATE. Use RC.open() for a buffered one
    # We could return the .raw from requests but this is seekable and
    # we more closely control the requests from rclone

    def __init__(self, remotefile, rc):
        self.rc = rc
        self.remotefile = (self.fs, self.remote) = rcpathsplit(remotefile)
        self.offset = 0

        self._head = head = rc._http_head(self.remotefile)
        if mx := head.get("Content-Length", None):
            self.maxsize = int(mx)
        else:
            self.maxsize = None

    def seekable(self):
        return True

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            self.offset = offset
        elif whence == io.SEEK_CUR:
            self.offset += offset
        elif whence == io.SEEK_END:
            if self.maxsize is None:
                raise io.UnsupportedOperation("Could not determine size")
            self.offset = self.maxsize + offset
        else:
            raise io.UnsupportedOperation()
        return self.offset

    def tell(self):
        return self.offset

    def readable(self):
        return True

    def writable(self):
        return False

    def read(self, size=-1, /):
        # Handle when reading everything since this is a lot faster
        if size >= 0:
            return super().read(size)

        if self.maxsize and self.offset >= self.maxsize:
            return b""

        return self.rc.read(
            self.remotefile,
            start=self.offset,
            end=None,
        )

    def readall(self):
        return self.read(-1)

    def readinto(self, b):
        if self.maxsize and self.offset >= self.maxsize:
            return 0

        N = len(b)
        try:
            chunk = self.rc.read(
                self.remotefile,
                start=self.offset,
                end=self.offset + N - 1,  # -1 since end is inclusive
            )
        except ValueError:
            b.clear()
            return 0

        n = len(chunk)
        self.offset += n
        if n != N:  # We know we hit the end since it returned less than we wanted
            self.maxsize = self.offset
        b[:n] = chunk
        return n


def rcpathsplit(path):
    """
    Splits the fs and the remote while acounting for special remotes and connection
    strings. If given as a 2-tuple or 2-list, will just return the value.

        rcpathsplit('single-file.ext') = ('./', 'single-file.ext')
        rcpathsplit('local/file.ext') = ('local', 'file.ext')
        rcpathsplit('remote:file.ext') = ('remote:', 'file.ext')
        rcpathsplit('remote:sub/file.ext') = ('remote:', 'sub/file.ext')
        rcpathsplit('remote:/sub/file.ext') = ('remote:', '/sub/file.ext')
        rcpathsplit(':http:sub/file.ext') = (':http:', 'sub/file.ext')
        rcpathsplit(":http,url='https://example.com':path/to/dir") = (":http,url='https://example.com':", 'path/to/dir')
        rcpathsplit(":http,url='https://example.com':path/t'o/dir/with'quote") = (":http,url='https://example.com':", "path/t'o/dir/with'quote")

    The algorithim is heuristic but should account for most cases. It removes the
    leading ':' if present (from an on-the-fly remote). Then it removes everything that
    is quoted. Afterwards, it splits at : and replaces the removed values. It does not
    require quotes to be matched anywhere but in the remote name but that is also an
    rclone requirement
    """
    if isinstance(path, (tuple, list)) and len(path) == 2:
        return path

    if getattr(path, "_is_rc_file", False):
        return path.remotefile

    path = str(path)

    if otf := path.startswith(":"):  # on the fly
        path = path[1:]

    reQUOTE = re.compile(
        r"""
         '{3}(?:[^\\]|\\.|\n)+?'{3}        # 3 single ticks
        |\"{3}(?:[^\\]|\\.|\n)+?\"{3}      # 3 double ticks
        |\".+?\"                           # 1 double tick
        |'.+?'                             # 1 single tick
        """,
        flags=re.MULTILINE | re.DOTALL | re.UNICODE | re.VERBOSE,
    )  # Regex to capture quoted statements

    reps = defaultdict(randstr)
    unquoted = reQUOTE.sub(lambda m: reps[m.group(0)], path)

    split = unquoted.split(":", 1)
    if len(split) == 1:  # Local path
        fs, remote = os.path.split(split[0])
        if not fs:
            fs = "./"
    else:
        fs, remote = split
        fs = f"{fs}:"
        if otf:
            fs = f":{fs}"

    for a, b in reps.items():
        fs = fs.replace(b, a)
        remote = remote.replace(b, a)
    return fs, remote


def rcpathjoin(*args):
    """
    This is like os.path.join but does some rclone-specific things because
    there could be a ':' in the first part.

    The second argument could be '/file', or 'file' and the first could have a colon.
        RcloneCLI.pathjoin('a','b')   # a/b
        RcloneCLI.pathjoin('a:','b')  # a:b
        RcloneCLI.pathjoin('a:','/b') # a:/b # Note that these are unlike os.path.join
        RcloneCLI.pathjoin('a','/b')  # a/b

    """
    args = [str(a) for a in args]  # Pathlib

    if len(args) <= 1:
        return "".join(args)

    root, first, rest = args[0], args[1], args[2:]

    if root.endswith("/"):
        root = root[:-1]

    if root.endswith(":") or first.startswith("/"):
        path = root + first
    else:
        path = f"{root}/{first}"

    path = os.path.join(path, *rest)
    return path


def random_port():
    with socket.socket() as sock:
        sock.bind(("", 0))
        return sock.getsockname()[1]


def dtfun(elapsed, dtmin=0.2, dtmax=1.5, k=0.15):
    """
    Standard logistic curve that is shifted into [dtmin,dtmax]
    with factor k
    """
    e = 2.718281828459045
    L = 1 / (1 + e ** (-k * elapsed))  # [0.5,1]
    L = (L - 0.5) / 0.5  # [0,1]
    L = (dtmax - dtmin) * L + dtmin  # [dtmin,dtmax]
    return L


def filter_cli2params(flags):
    """
    Converts and maps rclone flags into a _filter dict for rclone rc calls.

    Arguments:
        flags
            List or string of flags. If string, will apply shlex.split on them

    Returns:
        _filter dictionary

    Valid Flags:
        --delete-excluded
        --filter --filter-from
        --exclude --exclude-from
        --include --include-from
        --exclude-if-present
        --files-from --files-from-raw
        --min-age --max-age
        --min-size --max-size
        --ignore-case
        --metadata-filter --metadata-filter-from
        --metadata-exclude --metadata-exclude-from
        --metadata-include --metadata-include-from
    """
    if isinstance(flags, str):
        flags = shlex.split(flags)

    param2flag = {
        "DeleteExcluded": "--delete-excluded",
        "FilterRule": "--filter",
        "FilterFrom": "--filter-from",
        "ExcludeRule": "--exclude",
        "ExcludeFrom": "--exclude-from",
        "IncludeRule": "--include",
        "IncludeFrom": "--include-from",
        "ExcludeFile": "--exclude-if-present",
        "FilesFrom": "--files-from",
        "FilesFromRaw": "--files-from-raw",
        "MinAge": "--min-age",
        "MaxAge": "--max-age",
        "MinSize": "--min-size",
        "MaxSize": "--max-size",
        "IgnoreCase": "--ignore-case",
        "meta_FilterRule": "--metadata-filter",
        "meta_FilterFrom": "--metadata-filter-from",
        "meta_ExcludeRule": "--metadata-exclude",
        "meta_ExcludeFrom": "--metadata-exclude-from",
        "meta_IncludeRule": "--metadata-include",
        "meta_IncludeFrom": "--metadata-include-from",
    }

    BOOL_PARAMS = {"IgnoreCase", "DeleteExcluded"}
    SINGLE_PARAMS = {"MinAge", "MaxAge", "MinSize", "MaxSize"}

    parser = ThrowingArgumentParser()
    for param, flag in param2flag.items():
        if param in BOOL_PARAMS:
            parser.add_argument(flag, dest=param, action="store_true", default=None)
        elif param in SINGLE_PARAMS:
            parser.add_argument(flag, dest=param, default=None)
        else:
            parser.add_argument(
                flag,
                dest=param,
                action="append",
                default=None,
                metavar="VAL",
            )

    args = parser.parse_args(flags)

    params = {}
    meta_params = {}

    for param in param2flag:
        if not (val := getattr(args, param)):
            continue

        if param.startswith("meta_"):
            meta_params[param.removeprefix("meta_")] = val
        else:
            params[param] = val

    if meta_params:
        params["MetaRules"] = meta_params

    return params


def popen_streamer(proc, allow_error=False):
    """
    Takes a subprocess.Popen object and yields stdout and stderr
    simultaneously

    The proc must be opened with
        stdout=subprocess.PIPE,stderr=subprocess.PIPE

    Note that if the underlying process doesn't flush the output,
    it may not happen as expected.

    if PIPE is False, tells this function that subprocess if being
    written to files and will read them with a poll time of 1/100th
    of a second.

    Yields:
        ('stdout' or 'stderr', line)
    """

    Q = Queue()

    def _reader(oe):
        file = getattr(proc, oe)
        with file:
            for line in file:
                Q.put((oe, line))

        Q.put((oe, None))  # Done

    outthread = Thread(target=_reader, args=("stdout",))
    errthread = Thread(target=_reader, args=("stderr",))
    outthread.start(), errthread.start()

    c = 0
    while c < 2:
        oe, line = Q.get()
        if line is None:
            c += 1
            Q.task_done()
            continue
        yield oe, line
        Q.task_done()

    proc.wait()  # Should be done executing already
    if not allow_error:
        check_returncode(proc)

    Q.join()
    outthread.join()
    errthread.join()


FILTER_FLAGS = frozenset(
    {
        "--delete-excluded",
        "--exclude",
        "--exclude-from",
        "--exclude-if-present",
        "--files-from",
        "--files-from-raw",
        "--filter",
        "--filter-from",
        "--ignore-case",
        "--include",
        "--include-from",
        "--max-age",
        "--max-size",
        "--metadata-exclude",
        "--metadata-exclude-from",
        "--metadata-filter",
        "--metadata-filter-from",
        "--metadata-include",
        "--metadata-include-from",
        "--min-age",
        "--min-size",
    }
)

IGNORED_FILE_DATA = frozenset(
    [
        "IsDir",
        "Name",
        "ID",
        "Tier",
    ]
)



import datetime
import logging
import os

from .configuration import LOCK
from .utils import time2all, rpath2apath, apath2rpath
from .timestamps import timestamp_parser

logger = logging.getLogger(__name__)

__version__ = "20250328.0BETA"
MIN_RCLONE = 1, 63, 0


# This is a bit of a hack. When setup.py is run, if and only if, it is a git repo, this
# will get set. Otherwise, it is None
__git_version__ = {'version': '970e48a', 'origin': 'git@github.com:Jwink3101/dfb.git'}

_override_ts = os.environ.get("DFB_OVERRIDE_TIMESTAMP", None)
_override_unix = os.environ.get("DFB_OVERRIDE_UNIXTIME", None)
_override_offset = 0


def nowfun():
    """
    Return current datetime.
    This is its own function so _override_unix and _override_offset can
    be set in tests
    """
    if _override_ts:
        now = timestamp_parser(_override_ts)
    elif _override_unix:
        now = datetime.datetime.fromtimestamp(_override_unix, datetime.timezone.utc)
    else:
        now = datetime.datetime.now()

    if _override_offset:
        now += datetime.timedelta(seconds=_override_offset)

    return time2all(now)


def nowfun_obj():
    return nowfun().obj


# Monkey patch this into timestamps. We do this so that timestamps *can* exists 100%
# without dfb (such as if I ever want to copy/paste the code) but for dfb, we want this
# version of nowfun since it can be overridden for testing
from . import timestamps

timestamps.nowfun = nowfun

# For testing only. I can add fail points
_FAIL = set()



import os, sys
import argparse
import logging

__version__ = "20240102.0"

from dfb import __version__ as dfb_version
from dfb.cli import ISODATEHELP

from .link import dfblink

epilog = """\
Note on Mounting
----------------
The 'mount' location can point to a sub directory of the backup but the rclone
mount should be at the highest level. This is to allow for reference files to
correctly resolve.

For example, if you backups to 'backups:' and want to look at the subdirectory
'Documents/Pictures', you should

    $ rclone mount backups: mountpoint [mount flags ...]
    $ %(prog)s mountpoint/Documents/Pictures dest [link flags ...]

so that references point above 'Documents/Pictures' will resolve.

It is also suggested to use mount flags '--read-only' and maybe '--vfs-cache-mode full'
"""

description = """\
Make symlinks to an rclone mount (or local path) of the backups
"""


def cli(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "mount",
        metavar="mountpoint",
        help="""
            Specify where rclone is mounted including possible subdirectories. 
            See "Note on Mounting" for details on subdirectories
        """,
    )
    parser.add_argument("dest", help="Specify the destination directory for the links")

    parser.add_argument(
        "--before",
        "--at",
        default=None,
        help=f"Specify a timestamp to mimic. Defaults to latest. {ISODATEHELP}",
    )
    parser.add_argument(
        "--after",
        default=None,
        help="""
            Specify the earliest timestamp to include. Defaults to earliest. 
            See --before for details on timestamps
        """,
    )
    parser.add_argument(
        "--max-depth",
        default=None,
        type=int,
        metavar="N",
        help="""
            Specify a maximum depth to descend and create links. 
            Default is no limit. Top level is 0
            """,
    )
    parser.add_argument(
        "--allow-non-empty",
        action="store_true",
        help="Allow 'dest' to not be empty. Otherwise will error",
    )
    parser.add_argument(
        "--force-overwrite",
        action="store_true",
        help="Allow %(prog)s to overwrite existing files in 'dest'. Otherwise, will error if one already exists",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Verbosity. Specify twice for debug mode",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"dfb-link.{__version__},dfb-{dfb_version}",
    )

    args = parser.parse_args(argv)

    args.verbose = max([0, min([args.verbose, 2])])
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=[logging.WARNING, logging.INFO, logging.DEBUG][
            getattr(args, "verbose", 0)
        ],
    )

    logging.logger.debug(" argv: %s", str(argv))
    logging.logger.debug(" args: %s", args)

    try:
        dfblink(
            args.mount,
            args.dest,
            before=args.before,
            after=args.after,
            maxdepth=args.max_depth,
            allow_non_empty=args.allow_non_empty,
            force_overwrite=args.force_overwrite,
        )
    except Exception as E:
        logging.error(str(E))
        if args.verbose >= 2:
            raise
        sys.exit(2)



import os, sys
import argparse
import logging
from textwrap import dedent
from functools import partial

from . import __version__, __git_version__

logger = logging.getLogger(__name__)

_TESTMODE = False

DFB_CONFIG = os.environ.get("DFB_CONFIG", os.environ.get("DFB_CONFIG_FILE", None))
argv = None

ISODATEHELP = """
    Specify a date and timestamp in an ISO-8601 like format (YYYY-MM-DD[T]HH:MM:SS) with
    or without spaces, colons, dashes, "T", etc. Can optionally
    specify a numeric time zone (e.g. -05:00) or 'Z'. If no timezone is specified, 
    it is assumed *local* time. Alternatively, can specify unix time with a preceding
    'u' (e.g. 'u1678560662'). Or can specify a time difference from the current
    time with any (and only) of the following: second[s], minute[s], hour[s], day[s], 
    week[s]. Example: "10 days 1 hour 4 minutes 32 seconds". (The order doesn't matter). 
    Can also specify "now" for the current time.
    """


# This lets me control how argparse exits.
# https://stackoverflow.com/a/14728477
class ThrowingArgumentParserError(Exception):
    pass


class ThrowingArgumentParser(argparse.ArgumentParser):
    """Like regular argument parser but throws an exception"""

    def error(self, message):
        self.print_usage(sys.stderr)
        args = {"prog": self.prog, "message": message}
        msg = ("%(prog)s: error: %(message)s\n") % args
        raise ThrowingArgumentParserError(msg)


def parse(argv=None, shebanged=False):
    if argv is None:
        argv = sys.argv[1:]
    # Hacked commands
    if argv and argv[0] == "help":
        argv[0] = "--help"
    if argv and argv[0] == "version":
        argv[0] = "--version"

    # config_global_group = global_parent.add_argument_group(title="Config Settings")
    config_global = argparse.ArgumentParser(add_help=False)
    config_global_group = config_global.add_argument_group(
        title="Config & Cache Settings",
    )
    config_help = f"""
        (Required) Specify config file. Can also be specified via the 
        $DFB_CONFIG environment variable or is implied if executing the config
        file itself. $DFB_CONFIG is currently 
        {('set to ' + repr(DFB_CONFIG)) if DFB_CONFIG else 'not set'}. 
        """
    if shebanged:
        config_help += f" Currently implied as {shebanged!r}."
    config_global_group.add_argument(
        "--config",
        metavar="file",
        help=config_help,
        default=DFB_CONFIG,
        required=not bool(DFB_CONFIG),
    )
    config_global_group.add_argument(
        "-o",
        "--override",
        action="append",
        default=list(),
        metavar="'OPTION = VALUE'",
        help=(
            "Override any config option for this call only. Must be specified as "
            "'OPTION = VALUE', where VALUE should be proper Python (e.g. quoted strings). "
            """Example: --override "compare = 'mtime'". """
            "Override text is evaluated before *and* after the config file however, "
            "the variables 'pre' and 'post' are defined as True or False if it is "
            "before or after the config file. These can be used with conditionals to "
            "control overrides. See readme for details. "
            "Can specify multiple times. There is no input validation so do not specify "
            "untrusted inputs."
        ),
    )
    global_parent = argparse.ArgumentParser(add_help=False)
    global_group = global_parent.add_argument_group(
        title="Global Settings",
        description="Default verbosity is 1 for backup/restore/prune and 0 for listing",
    )
    global_group.add_argument(
        "-v", "--verbose", "--debug", action="count", help="+1 verbosity", default=0
    )
    global_group.add_argument(
        "-q", "--quiet", action="count", help="-1 verbosity", default=0
    )
    global_group.add_argument(
        "--temp-dir", help="Specify a temp dir. Otherwise will use Python's default"
    )

    when_parent = argparse.ArgumentParser(add_help=False)
    when_group = when_parent.add_argument_group(
        title="Time Specification", description=f"All TIMESTAMPs: {ISODATEHELP}"
    )
    when_group.add_argument(
        "--at",
        "--before",
        dest="before",
        metavar="TIMESTAMP",
        help=f"""
            Timestamp at which to show the files. If not specified, will be the latest.
            Note that if '--after' is set, this will not be the full snapshot in time.
            """,
    )
    when_group.add_argument(
        "--after",
        metavar="TIMESTAMP",
        help=f"""
            Only show files after the specified time. Note that this means the '--at' will
            not be the full snapshot.
            """,
    )

    when_group.add_argument(
        "--only",
        metavar="TIMESTAMP",
        help=f"""
            Only show files AT the specified time. 
            Shortcut for '--before TIMESTAMP --after TIMESTAMP' since both are inclusive.
            Useful if the exact timestamp is known such as from the 'timestamps' command.
            """,
    )
    restore_parent = argparse.ArgumentParser(add_help=False, parents=[when_parent])
    restore_parent.add_argument(
        "--no-check",
        action="store_true",
        help="Disable rclone comparing the source and the dest. If set, will restore everything",
    )

    exe_shell_parent = argparse.ArgumentParser(add_help=False)
    exe_dump_parent = argparse.ArgumentParser(add_help=False)

    exe_groups = []
    for par in (exe_shell_parent, exe_dump_parent):
        pargroup = par.add_argument_group(
            title="Execution Settings",
            description="Precedance follows the order specified in this help",
        )
        # dry_int_group = dry_int_group0.add_mutually_exclusive_group()
        pargroup.add_argument(
            "-n", "--dry-run", action="store_true", help="Do not execute any changes"
        )
        pargroup.add_argument(
            "-i",
            "--interactive",
            action="store_true",
            help="Display planned actions and prompt to continue or stop",
        )

        exe_groups.append(pargroup)

    exe_groups[0].add_argument(  # onto exe_shell_parent
        "--shell-script",
        metavar="FILE or -",
        help="""
            Rather than call rclone from within '%(prog)s', instead generate
            a shell script at %(metavar)s that will perform the same actions.
            This is useful to verify behavior and modify as needed. Note that
            this may not be perfect but should be close. Can specify "-" to print script
            """,
    )

    exe_groups[1].add_argument(  # onto exe_dump_parent
        "--dump",
        metavar="FILE or -",
        help="""
            ADVANCED USAGE. Will dump the JSONL data that represents the backup or prune. 
            This can be used to manually do the action and then with 
            `dfb advanced dbimport` to apply. Note that this is a more advanced
            form of --dry-run. If FILE ends in '.gz' or '.xz', it will be
            compressed respectively. 
            Can specify "-" to print the dump to stdout.
            """,
    )

    #### Main

    parser = ThrowingArgumentParser(
        description="dfb -- Date File Backup",
        prog="dfb",
        parents=[global_parent],
    )

    version = "%(prog)s-" + __version__
    if __git_version__:
        version += f"|{__git_version__['version']}"
    parser.add_argument("--version", action="version", version=version)

    subparsers = {}
    subparsers["main"] = subpar = parser.add_subparsers(
        dest="command",
        title="Commands",
        required=True,
        metavar="command",
        description="Run `%(prog)s <command> -h` for help",
    )

    #################################################
    ## Init
    #################################################

    init = subparsers["init"] = subpar.add_parser(
        "init",
        parents=[global_parent],
        help="write a new config file.",
    )
    init.add_argument(
        "--force-overwrite",
        action="store_true",
        help="Force %(prog)s to overwrite an existing config file",
    )
    init.add_argument(
        "config",
        metavar="config-file",
        help="Specify a config file destination",
    )

    #################################################
    ## Backup
    #################################################

    backup = subparsers["backup"] = subpar.add_parser(
        "backup",
        parents=[global_parent, config_global, exe_dump_parent],
        help="Run a backup",
    )
    backup.add_argument(
        "--subdir",
        default="",
        help="""
            Backup only a subset of the source. Move tracking (referencing) will not 
            work *outside* the subdir. It will look like a delete and then a later backup.
            WARNING: rclone will list the source with the subdir as the remote. 
            Filters that assume a path and/or are anchored to the root 
            (i.e. start with '/'), will NOT be applied correctly. 
            Use '--dry-run' or '--interactive' to verify! The variable 'subdir' is
            also defined in the config file which can be used with conditionals. 
            ⚠⚠⚠USE WITH CAUTION!⚠⚠⚠
            """,
    )
    backup.add_argument(
        "--refresh",
        action="store_true",
        help="""
            Refresh the local cache with a real listing of the remote destination. 
            This can be much slower as it must list all versions of all files
            however, it is useful if something has changed at the remote outside of
            %(prog)s (e.g., manual pruning). When used, will use 'remote_compare' attribute
            instead of 'compare'. This is the same as running `refresh` command but
            will list simultaneously.""",
    )
    backup.add_argument(
        "--refresh-use-snapshots",
        action=argparse.BooleanOptionalAction,
        dest="use_snapshots",
        default=True,
        help="""
            Whether or not to also download snapshots from the destination and
            update metadata. Note that the snapshots are _secondary_. They are
            not needed but enable src-to-src comparisons immediately after refresh
            and are faster for resolving references. Default: %(default)s.
            """,
    )

    #################################################
    ## Backup
    #################################################

    refresh = subparsers["refresh"] = subpar.add_parser(
        "refresh",
        parents=[global_parent, config_global],
        help="""
            Refresh the local cache with a real listing of the remote destination
            Same as calling backup with `--refresh` but
            can be used outside of a backup
            """,
    )
    refresh.add_argument(
        "--use-snapshots",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="""
            Whether or not to also download snapshots from the destination and
            update metadata. Note that the snapshots are _secondary_. They are
            not needed but enable src-to-src comparisons immediately after refresh
            and are faster for resolving references. Default: %(default)s.
            """,
    )

    #################################################
    ## restore-dir
    #################################################
    _h = "Restore a (sub)directory to a specified location"
    restore_dir = subparsers["restore-dir"] = subpar.add_parser(
        "restore-dir",
        aliases=["restore"],  # Need to reset below
        parents=[global_parent, config_global, restore_parent, exe_shell_parent],
        help=_h,
        description=_h,
    )
    restore_dir.add_argument(
        "--source-dir",
        dest="source",
        default="",
        help="Source directory. Default is the root",
    )

    # These aren't in restore_parent because (a) I need to control the order and (b)
    # the text is slightly different
    restore_dir.add_argument(
        "dest",
        help="""
        Destination directory. Can be a local destination (e.g. '/path/to/restore' or '.'), 
        an arbitrary rclone remote (e.g. myremote:restore/path) or relative
        to the configured source by specifying it as "@src" (e.g. @src/restore/path)
        """,
    )
    #################################################
    ## restore-file
    #################################################

    restore_file = subparsers["restore-file"] = subpar.add_parser(
        "restore-file",
        parents=[global_parent, config_global, restore_parent, exe_shell_parent],
        help="Restore a file to a specified location, file, or to stdout",
    )

    restore_file.add_argument(
        "source", help="File in the Backup. Optionally at the specified time"
    )

    restore_file.add_argument(
        "dest",
        help="""
        Destination directory or file (if --to). Can be a local destination 
        (e.g. '/path/to/restore' or '.'), 
        an arbitrary rclone remote (e.g. myremote:restore/path), relative
        to the configured source by specifying it as "@src" (e.g. @src/restore/path), or 
        specify as '-' to print to stdout.
        
        """,
    )

    restore_file.add_argument(
        "--to",
        action="store_true",
        help="""
        Assumes 'dest' is a file, not a directory. (i.e., uses 'rclone copyto' 
        instead of 'rclone copy')""",
    )
    #################################################
    ## File List
    #################################################
    list_parent = argparse.ArgumentParser(add_help=False, parents=[when_parent])

    list_parent.add_argument(
        "path", default="", nargs="?", help="Starting path. Defaults to the top"
    )

    # Setting for listings
    list_parent_settings = argparse.ArgumentParser(add_help=False)
    list_group = list_parent_settings.add_argument_group(title="Listing Settings")
    list_group.add_argument(
        "--header",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Print a header where applicable. Default: %(default)s",
    )
    list_group.add_argument(
        "--head",
        default=None,
        type=int,
        metavar="N",
        help="Include the first %(metavar)s lines plus --tail (if set).",
    )
    list_group.add_argument(
        "--tail",
        default=None,
        type=int,
        metavar="N",
        help="Include --head (if set) plus the last %(metavar)s lines.",
    )
    list_group.add_argument(
        "--human",
        action="store_true",
        help="Use human readable sizes",
    )
    list_group.add_argument(
        "--timestamp-local",
        action="store_true",
        help="""
            Specify timestamps in local time instead of UTC/Z (default). 
            Note, if applicable, all ModTimes are always local regardless""",
    )

    ls = subparsers["ls"] = subpar.add_parser(
        "ls",
        parents=[global_parent, list_parent, config_global, list_parent_settings],
        help="Cleanly list files and directories at the optionally specified time",
    )

    ls.add_argument(
        "-d",
        "--deleted",
        "--del",
        dest="deleted",
        action="count",
        default=0,
        help="""
            List deleted files with '<filename> (DEL)'. 
            Specify twice to ONLY include deleted files""",
    )
    ls.add_argument(
        "--full-path", action="store_true", help="Show full path when listing subdirs"
    )
    ls.add_argument(
        "-l",
        "--long",
        action="count",
        default=0,
        help="""
            Long listing with size, ModTime, path. 
            Specify twice for versions, total_size, size, ModTime, Timestamp, path.
            If a file is a reference, the size will be that of the referent with an "R"
            appended.
            """,
    )

    ls.add_argument(
        "--list",
        "--list-only",
        dest="list_only",
        choices=["files", "dirs", "both"],
        default=None,
        help="""
            Only list files or directories (or dirs). Default 'both' normally or 
            'files' for --recursive mode.
            """,
    )

    ls.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="""
            List all items recursively
            """,
    )

    ls.add_argument(
        "--real-path",
        "--rpath",
        action="count",
        default=0,
        dest="rpath",
        help="""
            Print the relevant (based on time settings) real path (rpath) of file. 
            Specify one to print the real-path of a reference file and twice to print
            the referent. Multiple specifications only matter for reference files.
            """,
    )

    snap = subparsers["lsnaps"] = subpar.add_parser(
        "snapshot",
        parents=[global_parent, list_parent, config_global],
        help="""
            Recursively list the files in line-delimited JSON at the 
            optionally specified time
            """,
    )

    de_group = snap.add_mutually_exclusive_group()
    de_group.add_argument(
        "-d",
        "--deleted",
        "--del",
        dest="deleted",
        action="count",
        default=0,
        help="""
            List deleted files as well.  
            Specify twice to ONLY include deleted files""",
    )
    de_group.add_argument(
        "-e",
        "--export",
        action="store_true",
        help="Export mode. Includes _all_ entries, not just the final one",
    )
    snap.add_argument(
        "-O",
        "--output",
        help="""
        Specify an output file. Otherwise will print to stdout. If the file ends in .gz
        or .xz, will use the respective compression.""",
    )

    tree = subparsers["tree"] = subpar.add_parser(
        "tree",
        parents=[global_parent, list_parent, config_global],
        help="""
            Recursively list files in a tree
            """,
    )
    tree.add_argument(
        "-d",
        "--deleted",
        "--del",
        dest="deleted",
        action="count",
        default=0,
        help="""
            List deleted files as well.  
            Specify twice to ONLY include deleted files""",
    )

    tree.add_argument(
        "--max-depth",
        type=int,
        metavar="N",
        default=-1,
        help="""
            Specify depth. The original path is 1. Default is none
            """,
    )

    versions = subparsers["versions"] = subpar.add_parser(
        "versions",
        epilog="""Fields are [reference_count],size,mtime,timestamp,[real-path]. 
            mtime is local and snapshot will depend on setting. 
            Size will be "D" for deleted items and *end* in "R" for a reference file""",
        parents=[global_parent, config_global, list_parent_settings],
        help="Show all versions of a file.",
    )
    versions.add_argument("filepath", help="Path to the file of interest")

    versions.add_argument(
        "--ref-count", action="store_true", help="Include the reference count"
    )
    versions.add_argument(
        "--real-path",
        "--rpath",
        dest="real_path",
        action="count",
        default=0,
        help="Include *full* real path of the file. Specify twice to include the rclone path",
    )

    timestamps = subparsers["timestamps"] = subpar.add_parser(
        "timestamps",
        parents=[global_parent, config_global, list_parent_settings, when_parent],
        help="""
            List all timestamps in the backup or within specified range. 
            Note that this will include ones that were nominally pruned but 
            without all files""",
    )
    timestamps.add_argument(
        "path",
        default="",
        nargs="?",
        help="""
            Starting path. Defaults to the top. Specifying a path will also change the 
            stats to _only_ consider that path
            """,
    )
    # timestamps.add_argument('path')
    #################################################
    ## Pruning
    #################################################
    prune = subparsers["prune"] = subpar.add_parser(
        "prune",
        epilog="""
            Pruning takes into account reference files and delete markers that need to
            be kept. Note that after pruning, it may appear possible to restore older
            than the prune time but the results are very unlikely to be correct! It is
            due to files that are not-yet-modified or are referenced. The prune 
            algorithm may miss some delete makers that technically could be deleted but
            it is more efficient not to try to identify them.
        """,
        parents=[global_parent, config_global, exe_dump_parent],
        help="Prune older versions of the files",
    )
    prune.add_argument(
        "when",
        help=f"""
        Specify file modification prune time. The modification time of a file is when
        the *next* file was written and not the original timestamp. 
        {ISODATEHELP.strip()}""",
    )
    prune.add_argument(
        "-N",
        "--keep-versions",
        default=0,
        type=int,
        dest="N",
        help="""
            Specify number of versions to keep past the specified time. This can be used
            to prune versions only. For example, to keep only the last 10 versions, 
            do "prune now -N 10". Can also be combined with a date. For example, to keep 
            the last 4 versions older than 30 days, specify "prune '30 days' -N 4". 
            Can also specify negative numbers to shift forward in 
            time (advanced usage). 
            """,
        # Not keeping this in the docs but...
        # For example, to prune the oldest 5 versions, you can do
        # "prune u0 -N -6" (where you need to do one additional to account for the oldest).
    )
    prune.add_argument(
        "--subdir",
        default="",
        metavar="dir",
        help="""
            Prune only files in '%(metavar)s'. In order to ensure that references do not
            break, this is mostly just a filter of what will be pruned rather than a
            major performance enhancement.
            """,
    )

    #################################################
    ## Summary
    #################################################
    summary = subparsers["summary"] = subpar.add_parser(
        "summary",
        parents=[
            global_parent,
            # list_parent,
            config_global,
            when_parent,
            # list_parent_settings,
        ],
        help="Summary of files",
    )

    summary.add_argument(
        "path", default="", nargs="?", help="Starting path. Defaults to the top"
    )

    #################################################
    ## Advanced Subparser
    #################################################
    adv = subparsers["adv"] = subpar.add_parser(
        "advanced",
        help="Advanced functions. Run `%(prog)s advanced -h` for help",
    )

    subparsers["adv_main"] = adv_subpar = adv.add_subparsers(
        dest="command",
        title="Commands",
        required=True,
        metavar="command",
        description="Run `%(prog)s <command> -h` for help",
    )

    #################################################
    ## Advanced-Import
    #################################################

    dbimport = subparsers["dbimport"] = adv_subpar.add_parser(
        "dbimport",
        parents=[global_parent, config_global],
        help="Import an exported list",
        description="""
            [ADVANCED] Import file(s) and append to database. 
            Will overwrite any existing data if applicable. Note: Does *not* upload
            the import file lists to the remote as in a backup.
            """,
    )

    dbimport.add_argument(
        "files",
        nargs="*",
        help="""File(s) to import. Can be any rclone path including local. 
                Will automatically decompress .gz or .xz files""",
    )
    # This lets the user specify positional or flag arguments
    dbimport.add_argument(
        "--files",
        nargs="*",
        metavar="file",
        action="append",
        default=[],
        dest="files2",  # will be merged in later
        help="""File(s) to import. Can be any rclone path including local. 
                Will automatically decompress .gz or .xz files""",
    )

    dbimport.add_argument(
        "--dirs",
        nargs="*",
        metavar="dir",
        action="append",
        default=[],
        help="""Directories of files import. Can be any rclone path including local. 
                Will automatically decompress .gz or .xz files. Will always import
                files then directories""",
    )

    dbimport.add_argument(
        "--reset",
        action="store_true",
        help="Reset the DB before import. Call without files to *just* reset",
    )
    dbimport.add_argument(
        "--upload",
        action="store_true",
        help="""
            Uploads the imported file(s). Will put them in a subdirectory of the snapshots
            with the current time and label each file as 'N.{filename}'
            """,
    )

    #################################################
    ## Advanced prune path
    #################################################
    prunepath = subparsers["prunepath"] = adv_subpar.add_parser(
        "prune-file",
        parents=[global_parent, config_global, exe_dump_parent],
        help="Prune a specific file (real-path or rpath)",
        description="""
            [ADVANCED] Prune a specific "real-path" (or "rpath") from the database,
            optionally including all references to the path.
            """,
    )

    prunepath.add_argument(
        "rpath",
        help="""
            Specify rpath(s) ("real paths") to prune. These are the paths where the 
            file is actually stored such as what you see with `versions --real-path`            
            """,
        nargs="+",
    )

    prunepath.add_argument(
        "--error-if-referenced",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="""
            If true (default), will error if there are references to the provided
            path(s). If false, will *also* delete those references.  Default: %(default)s.
        """,
    )
    #################################################
    ## Advanced timestamp-filters
    #################################################
    tsfilt = subparsers["tsfilt"] = adv_subpar.add_parser(
        "timestamp-include-filters",
        parents=[global_parent, config_global, when_parent],
        help="Create rclone --include filters for a time range",
        description="""
            Given a range of times specified, create and print rclone 
            --include filters that can be used for other rclone operations 
            (e.g. ls, ncdu) on the destination. Note that the filters are not perfect
            and could possibly include additional items if they happen to have the same
            timestamp in the name
            """,
    )

    tsfilt.add_argument(
        "path",
        default="",
        nargs="?",
        help="""
            Starting path. Defaults to the top.
            """,
    )

    #################################################
    ## cli utils Subparser
    #################################################
    cliutil = subparsers["utils"] = subpar.add_parser(
        "utils",
        help="CLI utility functions. Run `%(prog)s utils -h` for help",
    )

    subparsers["utils_main"] = cliutil_subpar = cliutil.add_subparsers(
        dest="command",
        title="Commands",
        required=True,
        metavar="command",
        description="Run `%(prog)s <command> -h` for help",
    )

    #################################################
    ## cli utils apath2rpath
    #################################################

    a2r_parser = subparsers["apath2rpath"] = cliutil_subpar.add_parser(
        "apath2rpath",
        parents=[],
        help="Convert apparent path (apath) and date to a real path (rpath)",
        description="""
            Convert apparent path (apath) and date to a real path (rpath)
            """,
    )

    a2r_parser.add_argument(
        "--date",
        metavar="TIMESTAMP",
        help=f"Specify timestamp for the filenames. Default is current time. {ISODATEHELP}",
    )
    a2r_parser.add_argument(
        "-0",
        "--print0",
        action="store_true",
        help="Seperate multiple items with a null byte instead of newline",
    )

    a2r_parser.add_argument(
        "files",
        nargs="+",
        help="""
            Specify one or more files. If '-' is specified, will read stdin 
            (and automatically handle newlines or null-bytes).
            """,
    )

    #################################################
    ## cli utils apath2rpath
    #################################################

    r2a_parser = subparsers["rpath2apath"] = cliutil_subpar.add_parser(
        "rpath2apath",
        parents=[],
        help="Convert real path (rpath) an apparent path (apath) and date",
        description="""
            Convert real path (rpath) an apparent path (apath) and date. Returns data in
            JSONLines format with an ISO8601 date
            """,
    )

    r2a_parser.add_argument(
        "--timestamp-local",
        action="store_true",
        help="""
            Return timestamps in local time instead of UTC""",
    )

    r2a_parser.add_argument(
        "files",
        nargs="+",
        help="""
            Specify one or more files. If '-' is specified, will read stdin 
            (and automatically handle newlines or null-bytes).
            """,
    )

    #################################################
    ## DONE
    #################################################
    args = parser.parse_args(argv)
    args._argv0 = argv

    if getattr(args, "only", None):
        args.before = args.after = args.only

    return args


def clishebang(argv=None):
    """
    This will add the config and assume backup if it isn't specified.
    Generally, this works by catching a failure on the argument parser then
    adding 'backup' command. That emits an incorrect error so we *try* to avoid
    that by adding 'backup' if nothing else is specified. For example"

        ./config.py # Will add backup right away
        ./config.py --flag # will have erorr message but then work.

    It isn't perfect but reduces the number of messages.

    """
    if argv is None:
        argv = sys.argv[1:]
    if not argv:
        print("Must specify a config", file=sys.stderr)
        sys.exit(2)

    config = argv[0]
    argv = argv[1:] + ["--config", config]

    # If argv is only len-2, no command has been specified. Therefore, try to skip
    # the try/except below since it will emit cli help when not needed. Will still
    # happen if any backup flags are used. Not a big deal.
    if len(argv) == 2:
        argv.insert(0, "backup")

    try:
        cliconfig = parse(argv, shebanged=config)
    except ThrowingArgumentParserError as E:
        try:
            cliconfig = parse(["backup"] + argv)
        except ThrowingArgumentParserError as E:
            print(*E.args, file=sys.stderr)
            sys.exit(2)

    r = _cli(cliconfig)
    if _TESTMODE:
        return r


def cli(argv=None):
    try:
        cliconfig = parse(argv)
    except ThrowingArgumentParserError as E:
        print(*E.args, file=sys.stderr)
        sys.exit(2)
    r = _cli(cliconfig)
    if _TESTMODE:
        return r


def _cli(cliconfig):
    # Reset aliases. I wish this was how it always worked.
    if cliconfig.command == "restore":
        cliconfig.command = "restore-dir"

    from .configuration import Config

    verbosity = 0
    if cliconfig.command in {
        "backup",
        "refresh",
        "restore-dir",
        "restore-file",
        "prune",
        "prune-file",
        "dbimport",
    }:
        verbosity += 1
    verbosity += getattr(cliconfig, "verbose", 0) - getattr(cliconfig, "quiet", 0)
    verbosity = max([0, verbosity])

    # handle cli utils first since they do not need a config file.
    if cliconfig.command == "apath2rpath":
        from .cliutils import cli_apath2rpath

        return cli_apath2rpath(cliconfig)
    elif cliconfig.command == "rpath2apath":
        from .cliutils import cli_rpath2apath

        return cli_rpath2apath(cliconfig)

    try:
        add_params = {}
        add_params["subdir"] = getattr(cliconfig, "subdir", "")

        # config also sets logging
        config = Config(
            cliconfig.config,
            tmpdir=getattr(cliconfig, "temp_dir", None),
            verbosity=verbosity,
            add_params=add_params,
        )
    except Exception as E:
        logger.error(f"parse: {E}")
        sys.exit(2)

    try:
        config.cliconfig = cliconfig
        argv = cliconfig._argv0
        logger.debug(f"{argv = }")

        if cliconfig.command == "init":
            config._write_template(force=cliconfig.force_overwrite)
            print(f"New config in {cliconfig.config!r}")

            return

        config.parse(override_txt="\n".join(cliconfig.override))
        logger.debug(f"{cliconfig = }")

        if cliconfig.command == "_config":
            global _TESTMODE
            _TESTMODE = True
            return config

        ###########################################
        ## Call out to the actual workers
        ###########################################
        # This will handle the refresh on it's own so it can be concurrent
        if cliconfig.command == "backup":
            from .backup import Backup

            config._set_auto()

            back = Backup(config)  # Two steps so the object is initialized even
            back.run()  # ... if it fails
            return back

        elif cliconfig.command == "refresh":
            from .dstdb import DFBDST

            config._set_auto()

            DFBDST(config).reset(use_snapshots=cliconfig.use_snapshots)
            return config
        elif cliconfig.command == "dbimport":
            from .dstdb import DFBDST

            # Update CLI args
            cliconfig.files.extend(g for group in cliconfig.files2 for g in group)
            cliconfig.dirs = [g for group in cliconfig.dirs for g in group]

            DFBDST(config).dbimport(
                cliconfig.files,
                cliconfig.dirs,
                reset=cliconfig.reset,
                upload=cliconfig.upload,
            )
            return config
        elif cliconfig.command == "snapshot":
            from .listing import snapshot

            snapshot(config)
            return config
        elif cliconfig.command == "tree":
            from .listing import tree

            tree(config)
            return config
        elif cliconfig.command == "ls":
            from .listing import ls

            ls(config)
            return config
        elif cliconfig.command == "versions":
            from .listing import file_versions

            file_versions(config)
            return config
        elif cliconfig.command == "timestamps":
            from .listing import timestamps

            timestamps(config)

        elif cliconfig.command in ("restore-dir", "restore-file"):
            from .restore import Restore

            return Restore(config)

        elif cliconfig.command in ("prune", "prune-file"):
            from .prune import Prune

            prune = Prune(config)
            if cliconfig.command == "prune":
                prune.bydate()
            else:
                prune.byrpaths()
            return prune
        elif cliconfig.command == "summary":
            from .listing import summary

            summary(config)

        elif cliconfig.command == "timestamp-include-filters":
            from .listing import timestamp_include_filters

            timestamp_include_filters(config)
            return config
        else:
            logger.error(f"Unrecognized command {cliconfig.command!r}")
            return config

    except Exception as E:
        logger.error("")
        logger.error(str(E))
        logger.error("")

        try:
            # Call fail_shell iff cliconfig.command == "backup"
            if cliconfig.command == "backup" and config.fail_shell:
                from .utils import shell_runner

                logger.info("Running 'fail_shell' commands. Note: May not get logged")
                env = {
                    "LOGPATH": str(log.log_file.resolve()),
                    "DEBUGPATH": str(log.debug_file.resolve()),
                    "CONFIGDIR": os.path.dirname(config._configpath),
                }
                shell_runner(
                    config.fail_shell,
                    dry=config.cliconfig.dry_run,
                    env=env,
                    prefix="fail",
                )
            if cliconfig.command == "backup":
                logger.info(
                    "Will attempt to save logs and/or snapshots if configured. May fail"
                )
                back.upload_logs()
        except:
            logging.error("Saving logs and running fail_shell didn't work")

        if config.verbosity > 1:
            raise

        if not _TESTMODE:
            sys.exit(1)
    finally:
        try:
            config.rc.stop()
        except:
            pass



#!/usr/bin/env python
import logging
import os

from errno import EACCES
from os.path import realpath
from threading import Lock

from .fuse import FUSE, FuseOSError, Operations, LoggingMixIn

###########################################
import os, sys, json, time
from functools import lru_cache
from threading import Thread

p = os.path.abspath(os.path.dirname(__file__))
if p not in sys.path:
    sys.path.insert(0, p)

from dfb import __version__
from dfb.cli import ISODATEHELP
from dfb.timestamps import timestamp_parser
from dfb.dstdb import rpath2apath as _rpath2apath, NoTimestampInNameError

from collections import defaultdict


def rpath2apath(rpath):
    try:
        return _rpath2apath(rpath)
    except (NoTimestampInNameError, ValueError):
        return rpath, None, ""


REF_IS_V1 = ".REF_IS_V1"


class DFBMount:
    def __init__(
        self,
        ts=None,
        remove_empty=False,
        use_cache=False,
        cache_reset_min=None,
    ):
        self.ts = ts
        self.remove_empty = remove_empty
        if use_cache:
            maxsize = 128
            self.listdir = lru_cache(maxsize=maxsize)(self._listdir)
            self.apath2rpath = lru_cache(maxsize=8 * maxsize)(self._apath2rpath)
            if cache_reset_min:
                Thread(
                    target=self._reset_cache, args=(cache_reset_min,), daemon=True
                ).start()
        else:
            self.listdir = self._listdir
            self.apath2rpath = self._apath2rpath

    def _reset_cache(self, cache_reset_min):
        while True:
            time.sleep(cache_reset_min * 60)
            self.listdir.cache_clear()
            self.apath2rpath.cache_clear()

    def _listdir(self, dirpath, _empty_check=False, return_map=False):
        print(f"_listdir {dirpath = }")
        subdirs = []
        subfiles = []
        for item in os.listdir(dirpath):
            item = os.path.join(dirpath, item)
            {True: subdirs, False: subfiles}[os.path.isdir(item)].append(item)

        items = defaultdict(list)

        for item in subfiles:
            aname, ts, flag = rpath2apath(item)
            if self.ts and ts and ts > self.ts:
                continue

            if flag == "R":
                with open(item, "rt") as fp:
                    referent = fp.read().strip()

                # Handle different versions here
                try:
                    referent = json.loads(referent)
                except json.JSONDecodeError:
                    referent = {"ver": 1, "path": referent}

                ver = referent["ver"]
                if ver == 1:
                    item = item + REF_IS_V1
                    flag = ""
                elif ver == 2:
                    path = os.path.join(os.path.dirname(item), referent["rel"])
                    path = os.path.normpath(path)
                    item = path

            # keep ts so we sort by ts, not item in case of reference
            items[aname].append((ts, item, flag))

        # end for item in subitems

        for val in items.values():
            val.sort()

        items = {k: v[-1] for k, v in items.items()}
        items = {k: rpath for k, (ts, rpath, flag) in items.items() if flag != "D"}

        if _empty_check and items:
            return True

        for item in subdirs:
            if self.remove_empty:
                # Do a recursive call to listdir with _empty_check=True to get
                # results ASAP. (It will return as soon as the file is found)

                try:
                    ssub = self.listdir(item, _empty_check=True)
                except PermissionError:
                    ssub = True  # Assume it's not empty and move on

                if not ssub:  # Empty
                    continue
                elif _empty_check:
                    return True  # do not keep going deeper
            items[item] = item

        if return_map:
            return items
        return list(items)

    def _apath2rpath(self, apath):
        print(f"_apath2rpath {apath = }")
        dirpath = os.path.dirname(apath)
        items = self.listdir(dirpath, return_map=True)
        try:
            return items[apath]
        except KeyError:
            return apath


class DFBLoop(LoggingMixIn, Operations):
    def __init__(
        self,
        root,
        ts=None,
        remove_empty=False,
        use_cache=False,
        cache_reset_min=None,
    ):
        self.dfbmount = DFBMount(
            ts=ts,
            remove_empty=remove_empty,
            use_cache=use_cache,
            cache_reset_min=cache_reset_min,
        )

        self.root = realpath(root)
        self.rwlock = Lock()

    def __call__(self, op, apath, *args):
        # This is *just* a passthrough so keep it as apath!
        return super(DFBLoop, self).__call__(
            op, os.path.join(self.root, apath.removeprefix("/")), *args
        )

    def access(self, apath, mode):
        rpath = self.dfbmount.apath2rpath(apath)
        if not os.access(rpath, mode):
            raise FuseOSError(EACCES)

    #     chmod = os.chmod
    #     chown = os.chown
    #
    #     def create(self, path, mode):
    #         return os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode)
    #
    #     def flush(self, path, fh):
    #         return os.fsync(fh)
    #
    #     def fsync(self, path, datasync, fh):
    #         if datasync != 0:
    #             return os.fdatasync(fh)
    #         else:
    #             return os.fsync(fh)

    def getattr(self, apath, fh=None):
        rpath = self.dfbmount.apath2rpath(apath)
        st = os.lstat(rpath)
        keys = [
            "st_atime",
            "st_ctime",
            "st_gid",
            "st_mode",
            "st_mtime",
            "st_nlink",
            "st_size",
            "st_uid",
        ]
        return {k: getattr(st, k) for k in keys}

    getxattr = None

    #     def link(self, target, source):
    #         return os.link(self.root + source, target)

    listxattr = None

    #     mkdir = os.mkdir
    #     mknod = os.mknod
    #     open = os.open
    def open(self, apath, flags):
        rpath = self.dfbmount.apath2rpath(apath)
        return os.open(rpath, flags)

    def read(self, apath, size, offset, fh):
        rpath = self.dfbmount.apath2rpath(apath)
        with self.rwlock:
            if rpath.endswith(REF_IS_V1):
                rpath = rpath.removesuffix(REF_IS_V1)
                print("WARNING: V1 style reference")
            with open(rpath, "rb") as fp:
                fp.seek(offset)
                return fp.read(size)  #

    #             os.lseek(fh, offset, 0)
    #             return os.read(fh, size)

    def readdir(self, path, fh):
        res = self.dfbmount.listdir(path)
        res = [os.path.basename(r) for r in res]
        res.sort()
        return [".", ".."] + res

    #     readlink = os.readlink

    #     def release(self, path, fh):
    #         return os.close(fh)

    #     def rename(self, old, new):
    #         return os.rename(old, self.root + new)
    #
    #     rmdir = os.rmdir

    def statfs(self, apath):
        rpath = self.dfbmount.apath2rpath(apath)
        stv = os.statvfs(rpath)
        return dict(
            (key, getattr(stv, key))
            for key in (
                "f_bavail",
                "f_bfree",
                "f_blocks",
                "f_bsize",
                "f_favail",
                "f_ffree",
                "f_files",
                "f_flag",
                "f_frsize",
                "f_namemax",
            )
        )


#     def symlink(self, target, source):
#         return os.symlink(source, target)
#
#     def truncate(self, path, length, fh=None):
#         with open(path, 'r+') as f:
#             f.truncate(length)
#
#     unlink = os.unlink
#     utimens = os.utime
#
#     def write(self, path, data, offset, fh):
#         with self.rwlock:
#             os.lseek(fh, offset, 0)
#             return os.write(fh, data)

epilog = """\
Note on Mounting
----------------
This is a thin wrapper around a simple loop-back file system pointed to the dfb
destination. RcloneCLI's mounting capabilities is significantly more powerful so this
lets rclone manage the files and just wraps that mount. This is also read-only so it
makes sense to also have the rclone mount be read-only.

Example: Assume `dst = "backups:server"`

    $ rclone mount backups:server rclone_mount --vfs-cache-mode full --read-only
    $ %(prog)s rclone_mount dfb_mount
   
Note that dfb-mount will work with multiple directories. For example, you have more than
just 'server' on 'backups:', you can rclone mount just 'backups:' then point %(prog)s 
to the top level.
"""

description = """\
dfb FUSE thin-wrapper mount. Must FIRST mount the destination with rclone. Does **not**
work with libfuse3. Must use libfuse2 on Linux
"""


def cli(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("rclone_mount_dest", help="Specify where rclone is mounted")
    parser.add_argument("mount_dest", help="Mount dest")
    parser.add_argument(
        "--before",
        "--at",
        help=f"Specify a timestamp. {ISODATEHELP}",
        dest="ts",
    )
    ## TODO: --after flag
    parser.add_argument(
        "--remove-empty-dirs",
        action="store_true",
        help="""
            Try to determine if a directory is empty. Due to dfb keeping all copies
            of files, it is possible to have directories with no active files. This will
            try to determine that. It is disabled by default because of the performance
            hit of having to resurse until files are found""",
    )
    parser.add_argument(
        "--cache",
        action="store_true",
        help="""\
            Whether to cache dir listings. If this is going to be long-running, do not 
            use since new backups won't be updated. But useful for short jobs.
            """,
    )
    parser.add_argument(
        "--cache-reset",
        metavar="minutes",
        type=float,
        help="Specify an interval, in minutes, to reset the cache. If not set, cache will not reset",
    )
    parser.add_argument(
        "--version", action="version", version=f"dfb-mount_dbf.{__version__}"
    )

    args = parser.parse_args(argv)
    dfb = DFBMount()
    # logging.basicConfig(level=logging.DEBUG)
    if args.ts:
        args.ts = timestamp_parser(args.ts, aware=True, epoch=True)
    fuse = FUSE(
        DFBLoop(
            args.rclone_mount_dest,
            ts=args.ts,
            remove_empty=args.remove_empty_dirs,
            use_cache=args.cache,
            cache_reset_min=args.cache_reset,
        ),
        args.mount_dest,
        raw_fi=False,
        foreground=True,
        allow_other=True,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, shutil
from pathlib import Path
import gzip as gz
import subprocess
import json
from textwrap import dedent

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

from dfb.prune import BrokenReferenceError


# Local
import testutils

# testing
import pytest


def test_basic_cases():
    test = testutils.Tester(name="basic_cases")

    """
       init  del3  del5  del7   dm5   dm7   mix 
     1   *     *     *     *     *     *     *  
               │     │     │     │     │     │  
               │     │     │     │     │     │  
     3         D     │     │     *     *     D  
                     │     │     │     │        
                     │     │     │     │        
     5               D     │     D     *     *  
                           │           │     │  
                           │           │     │  
     7                     D           D     D   
     """

    # + tags=[]
    test.config["renames"] = False
    test.config["compare"] = "hash"
    test.write_config()
    vq = ["-q"]

    # + tags=[]
    test.write_pre("src/init.txt", "init")
    test.write_pre("src/del3.txt", "del3")
    test.write_pre("src/del5.txt", "del5")
    test.write_pre("src/del7.txt", "del7")
    test.write_pre("src/dm5.txt", "dm5")
    test.write_pre("src/dm7.txt", "dm7")
    test.write_pre("src/mix.txt", "mix")
    test.backup(*vq, offset=1)

    os.unlink("src/del3.txt")
    test.write_pre("src/dm5.txt", "dm5.3")
    test.write_pre("src/dm7.txt", "dm7.3")
    os.unlink("src/mix.txt")
    test.backup(*vq, offset=3)

    os.unlink("src/del5.txt")
    os.unlink("src/dm5.txt")
    test.write_pre("src/dm7.txt", "dm7.5")
    test.write_pre("src/mix.txt", "mix2")
    test.backup(*vq, offset=5)

    os.unlink("src/del7.txt")
    os.unlink("src/dm7.txt")
    os.unlink("src/mix.txt")
    test.backup(*vq, offset=7)

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u2", *vq, offset=9)
    assert set(prune.rpaths) == set()

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u4", *vq, offset=11)
    assert set(prune.rpaths) == {
        ("del3.19700101000001.txt", 4),
        ("del3.19700101000003D.txt", -1),
        ("dm5.19700101000001.txt", 3),
        ("dm7.19700101000001.txt", 3),
        ("mix.19700101000001.txt", 3),
    }
    # -

    prune = test.call("prune", "--dry-run", "u6", *vq, offset=13)
    assert set(prune.rpaths) == {
        ("del3.19700101000001.txt", 4),
        ("del3.19700101000003D.txt", -1),
        ("del5.19700101000001.txt", 4),
        ("del5.19700101000005D.txt", -1),
        ("dm5.19700101000001.txt", 3),
        ("dm5.19700101000003.txt", 5),
        ("dm5.19700101000005D.txt", -1),
        ("dm7.19700101000001.txt", 3),
        ("dm7.19700101000003.txt", 5),
        ("mix.19700101000001.txt", 3),
        ("mix.19700101000003D.txt", -1),  # Example of 2C delete
    }

    prune = test.call("prune", "--dry-run", "u8", *vq, offset=15)
    assert set(prune.rpaths) == {
        ("del3.19700101000001.txt", 4),
        ("del3.19700101000003D.txt", -1),
        ("del5.19700101000001.txt", 4),
        ("del5.19700101000005D.txt", -1),
        ("dm5.19700101000001.txt", 3),
        ("dm5.19700101000003.txt", 5),
        ("dm5.19700101000005D.txt", -1),
        ("dm7.19700101000001.txt", 3),
        ("dm7.19700101000003.txt", 5),
        ("mix.19700101000001.txt", 3),
        ("mix.19700101000003D.txt", -1),  # Example of 2C delete
    }.union(  # These are now removable
        {
            ("del7.19700101000001.txt", 4),
            ("del7.19700101000007D.txt", -1),
            ("dm7.19700101000005.txt", 5),
            ("dm7.19700101000007D.txt", -1),
            ("mix.19700101000005.txt", 4),
            ("mix.19700101000007D.txt", -1),
        }
    )

    # Now do it for real. They should result is a different group but they should build on the differences

    prune = test.call("prune", "u2", *vq, offset=17)

    # Should be the same
    prune = test.call("prune", "u4", *vq, offset=19)

    prune = test.call("prune", "u6", *vq, offset=21)
    assert set(prune.rpaths) == (
        {  # above's u6
            ("del3.19700101000001.txt", 4),
            ("del3.19700101000003D.txt", -1),
            ("del5.19700101000001.txt", 4),
            ("del5.19700101000005D.txt", -1),
            ("dm5.19700101000001.txt", 3),
            ("dm5.19700101000003.txt", 5),
            ("dm5.19700101000005D.txt", -1),
            ("dm7.19700101000001.txt", 3),
            ("dm7.19700101000003.txt", 5),
            ("mix.19700101000001.txt", 3),
            ("mix.19700101000003D.txt", -1),  # Example of 2C delete
        }
        - {
            ("del3.19700101000001.txt", 4),
            ("del3.19700101000003D.txt", -1),
            ("dm5.19700101000001.txt", 3),
            ("dm7.19700101000001.txt", 3),
            ("mix.19700101000001.txt", 3),
        }
    )

    prune = test.call("prune", "--dry-run", "u8", *vq, offset=23)
    assert set(prune.rpaths) == {
        ("del7.19700101000001.txt", 4),
        ("del7.19700101000007D.txt", -1),
        ("dm7.19700101000005.txt", 5),
        ("dm7.19700101000007D.txt", -1),
        ("mix.19700101000005.txt", 4),
        ("mix.19700101000007D.txt", -1),
    }

    # Test import with prune
    test.call("refresh")
    cmd = [
        "advanced",
        "dbimport",
        "-v",
        *sorted(str(p) for p in Path("dst/.dfb/snapshots/").glob("*.jsonl*")),
    ]
    test.call(*cmd)


def test_moves():
    """
        f0─┐  f1    f2   f3
     1   * └─┐
         │   └─┐
         │     │
     3   D     R


     5   * ───┐
         │    └───┐
         │        └───
     7   D           R
                     │
                     │
     9   *           *    *
         │           │
         │           │
    11   D           *
    """
    test = testutils.Tester(name="moves")

    # + tags=[]
    test.config["renames"] = "hash"
    test.config["compare"] = "hash"
    test.write_config()
    vq = ["-q"]

    # + tags=[]
    test.write_pre("src/f0.txt", "f0")
    test.backup(*vq, offset=1)

    test.move("src/f0.txt", "src/f1.txt")
    test.backup(*vq, offset=3)

    test.write_pre("src/f0.txt", "f0-2")
    test.backup(*vq, offset=5)

    test.move("src/f0.txt", "src/f2.txt")
    test.backup(*vq, offset=7)

    test.write_pre("src/f0.txt", "f0-3")
    test.write_pre("src/f2.txt", "mod")
    test.write_pre("src/f3.txt", "f3")
    test.backup(*vq, offset=9)

    os.unlink("src/f0.txt")
    test.write_pre("src/f2.txt", "mods")
    test.backup(*vq, offset=11)

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u2", *vq)
    assert set(prune.rpaths) == set()

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u4", *vq)
    assert set(prune.rpaths) == set()  # All are references

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u6", *vq)
    assert set(prune.rpaths) == set()  # All are STILL references

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u8", *vq)
    # No longer need  the D to block the ref since there is a new one
    assert set(prune.rpaths) == {("f0.19700101000003D.txt", -1)}

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u10", *vq)
    assert set(prune.rpaths) == {
        ("f2.19700101000007R.txt", 0),  # reference get deleted too
        ("f0.19700101000003D.txt", -1),  # From above
        # Now  blocked by 7D and no longer referecned by f2.7!
        ("f0.19700101000005.txt", 4),
    }

    # + tags=[]
    prune = test.call("prune", "--dry-run", "u12", *vq)
    assert set(prune.rpaths) == {
        ("f0.19700101000003D.txt", -1),  # From above
        ("f0.19700101000005.txt", 4),  # Now  blocked by 7D
        ("f0.19700101000007D.txt", -1),
        ("f0.19700101000009.txt", 4),
        ("f2.19700101000007R.txt", 0),
        ("f2.19700101000009.txt", 3),
    }
    # -

    # Useful to help visualize
    test.call("versions", "f0.txt", "--ref-count", "--real-path")
    test.call("versions", "f1.txt", "--ref-count", "--real-path")
    test.call("versions", "f2.txt", "--ref-count", "--real-path")
    test.call("versions", "f3.txt", "--ref-count", "--real-path")


def test_modes():
    test = testutils.Tester(name="modes")

    test.write_config()

    # +
    test.write_pre("src/file.txt", "1")
    test.backup("-q", offset=1)

    test.write_pre("src/file.txt", "1.3")
    test.backup("-q", offset=3)

    test.write_pre("src/file.txt", "1.3.5")
    test.backup("-q", offset=5)

    test.write_pre("src/file.txt", "1.3.7")
    test.backup("-q", offset=7)

    # +
    # **MANUAL** -- It works
    # test.call('prune','u6','-i')
    # -


#     prune = test.call("prune", "u6", "--shell-script", "-")
#     print(prune.rpaths)
#
#     prune = test.call("prune", "u6", "--shell-script", "prune.sh")
#     print(prune.rpaths)
#     with open("prune.sh") as fp:
#         print(fp.read())


def test_subdir():
    test = testutils.Tester(name="prune_subdir")
    test.write_config()

    vq = ["-q"]

    test.write_pre("src/nothing.txt", "do nothing")
    test.write_pre("src/mod.txt", "will mod .")
    test.write_pre("src/sub1/mod_sub.txt", "will mod in sub .")
    test.write_pre("src/sub2/move_at_5.txt", "will move")
    test.backup(offset=1)

    test.write_pre("src/mod.txt", "will mod ..")
    test.write_pre("src/sub1/mod_sub.txt", "will mod in sub ..")
    test.backup(offset=3)

    test.write_pre("src/mod.txt", "will mod ...")
    test.write_pre("src/sub1/mod_sub.txt", "will mod in sub ...")
    test.move("src/sub2/move_at_5.txt", "src/new/NEW.txt")
    test.backup(offset=5)

    test.write_post("src/new/NEW.txt", "neww")
    test.backup(offset=7)

    # Test it
    assert set(test.call("prune", "--dry-run", "u4", *vq).rpaths) == {
        ("mod.19700101000001.txt", 10),
        ("sub1/mod_sub.19700101000001.txt", 17),
    }
    assert set(test.call("prune", "--dry-run", "u4", "--subdir", "sub1").rpaths) == {
        ("sub1/mod_sub.19700101000001.txt", 17)
    }
    assert (
        set(test.call("prune", "--dry-run", "u6", "--subdir", "sub2", *vq).rpaths)
        == set()
    )

    prune = test.call("prune", "--dry-run", "u8", "--subdir", "sub2", *vq)
    assert set(prune.rpaths) == set()
    # WARNING this used to be
    #    ("sub2/move_at_5.19700101000001.txt", 9),
    #    ("sub2/move_at_5.19700101000005D.txt", -1),
    # but this is WRONG! Since we aren't pruning outside of subdir, those references
    # need to remain! This was a bug and a bad test

    prune = test.call("prune", "--dry-run", "u8", *vq)
    assert set(prune.rpaths) == {
        ("mod.19700101000003.txt", 11),
        ("sub1/mod_sub.19700101000001.txt", 17),
        ("mod.19700101000001.txt", 10),
        ("new/NEW.19700101000005R.txt", 0),
        ("sub1/mod_sub.19700101000003.txt", 18),
        ("sub2/move_at_5.19700101000001.txt", 9),
        ("sub2/move_at_5.19700101000005D.txt", -1),
    }


def test_disable():
    test = testutils.Tester(name="prune_dry")
    test.config["disable_prune"] = True
    test.write_config()

    test.write_pre("src/file.txt", "file")
    test.write_pre("src/to_del.txt", "del")
    test.backup(offset=1)

    test.write_post("src/file.txt", "file")
    os.unlink("src/to_del.txt")
    test.backup(offset=3)

    prune = test.call("prune", "--dry-run", "u4")
    assert set(prune.rpaths) == {
        ("to_del.19700101000001.txt", 3),
        ("to_del.19700101000003D.txt", -1),
        ("file.19700101000001.txt", 4),
    }
    assert os.path.exists("dst/to_del.19700101000001.txt")
    assert os.path.exists("dst/to_del.19700101000003D.txt")
    assert os.path.exists("dst/file.19700101000001.txt")

    prune = test.call("prune", "u4")
    assert set(prune.rpaths) == {
        ("to_del.19700101000001.txt", 3),
        ("to_del.19700101000003D.txt", -1),
        ("file.19700101000001.txt", 4),
    }
    assert os.path.exists("dst/to_del.19700101000001.txt")
    assert os.path.exists("dst/to_del.19700101000003D.txt")
    assert os.path.exists("dst/file.19700101000001.txt")

    prune = test.call("prune", "u4", "--override", "disable_prune = False")
    assert set(prune.rpaths) == {
        ("to_del.19700101000001.txt", 3),
        ("to_del.19700101000003D.txt", -1),
        ("file.19700101000001.txt", 4),
    }
    assert not os.path.exists("dst/to_del.19700101000001.txt")
    assert not os.path.exists("dst/to_del.19700101000003D.txt")
    assert not os.path.exists("dst/file.19700101000001.txt")

    # Prune-file
    test.call("advanced", "prune-file", "file.19700101000003.txt", "--dry-run")
    assert os.path.exists("dst/file.19700101000003.txt")

    test.call("advanced", "prune-file", "file.19700101000003.txt")
    assert os.path.exists("dst/file.19700101000003.txt")

    test.call(
        "advanced",
        "prune-file",
        "file.19700101000003.txt",
        "--override",
        "disable_prune = False",
    )
    assert not os.path.exists("dst/file.19700101000003.txt")


def test_basic_versions():
    test = testutils.Tester(name="prune_versions_basic")
    test.write_config()

    test.write_pre("src/mod_all.txt", "1")
    test.write_pre("src/nomod.txt", "do not touch")
    test.write_pre("src/del.txt", "del")
    test.write_pre("src/mv.txt", "mv")
    test.backup(offset=1)

    test.write_pre("src/mod_all.txt", "13")
    os.unlink("src/del.txt")
    test.move("src/mv.txt", "src/moved.txt")
    test.backup(offset=3)

    test.write_pre("src/mod_all.txt", "135")
    test.backup(offset=5)

    test.write_pre("src/mod_all.txt", "1357")
    test.backup(offset=7)

    assert set(test.call("prune", "now", "-n").rpaths) == {
        ("del.19700101000001.txt", 3),
        ("del.19700101000003D.txt", -1),
        ("mod_all.19700101000001.txt", 1),
        ("mod_all.19700101000003.txt", 2),
        ("mod_all.19700101000005.txt", 3),
    }

    assert set(test.call("prune", "now", "-N", "1", "-n").rpaths) == {
        ("mod_all.19700101000001.txt", 1),
        ("mod_all.19700101000003.txt", 2),
    }

    assert set(test.call("prune", "now", "-N", "2", "-n").rpaths) == {
        ("mod_all.19700101000001.txt", 1),
    }

    # Forward in time.
    assert set(test.call("prune", "u-1", "-N", "-2", "-n").rpaths) == {
        ("del.19700101000001.txt", 3),
        ("del.19700101000003D.txt", -1),
        ("mod_all.19700101000001.txt", 1),
    }

    # Make sure you can't go beyond the latest
    assert set(test.call("prune", "now", "-N", "-200", "-n").rpaths) == {
        ("del.19700101000001.txt", 3),
        ("del.19700101000003D.txt", -1),
        ("mod_all.19700101000001.txt", 1),
        ("mod_all.19700101000003.txt", 2),
        ("mod_all.19700101000005.txt", 3),
    }
    assert set(test.call("prune", "u0", "-N", "-200", "-n").rpaths) == {
        ("del.19700101000001.txt", 3),
        ("del.19700101000003D.txt", -1),
        ("mod_all.19700101000001.txt", 1),
        ("mod_all.19700101000003.txt", 2),
        ("mod_all.19700101000005.txt", 3),
    }


def test_prune_file():
    test = testutils.Tester(name="prune_file")
    test.config["rename_method"] = "reference"
    test.write_config()

    test.write_pre("src/mod.txt", ".")
    test.write_pre("src/keep.txt", "keep")
    test.write_pre("src/move0.txt", "mv0")

    test.backup(offset=1)

    test.write_post("src/mod.txt", "..")
    shutil.move("src/move0.txt", "src/move1.txt")

    test.backup(offset=3)

    test.write_post("src/mod.txt", "...")
    shutil.move("src/move1.txt", "src/move2.txt")

    test.backup(offset=5)

    # Check modes
    try:
        test.call(
            "advanced", "prune-file", "move0.19700101000001.txt", "-v", "--dry-run"
        )
        assert False
    except BrokenReferenceError:
        pass

    prune = test.call(
        "advanced",
        "prune-file",
        "move0.19700101000001.txt",
        "--dry-run",
        "--no-error-if-referenced",
    )
    assert prune.rpaths == {
        ("move1.19700101000003R.txt", 0),
        ("move0.19700101000001.txt", 3),
        ("move2.19700101000005R.txt", 0),
    }

    for path in (
        "move1.19700101000003R.txt",
        "move0.19700101000001.txt",
        "move2.19700101000005R.txt",
    ):
        assert (test.pwd / "dst" / path).exists()
    # Now do the prune for real and verify
    prune = test.call(
        "advanced",
        "prune-file",
        "move0.19700101000001.txt",
        "--no-error-if-referenced",
    )
    for path in (
        "move1.19700101000003R.txt",
        "move0.19700101000001.txt",
        "move2.19700101000005R.txt",
    ):
        assert not (test.pwd / "dst" / path).exists()


if __name__ == "__main__":
    test_basic_cases()
    test_moves()
    test_modes()
    test_subdir()
    test_disable()
    test_basic_versions()
    test_prune_file()

    print("=" * 50)
    print(" All Passed ".center(50, "="))
    print("=" * 50)



"""
Unit-like tests
"""

import os, sys, time

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

from dfb.utils import smart_splitext, time2all, head_tail_table, parse_bytes
from dfb.dstdb import rpath2apath, apath2rpath

DATED_SPLIT_TESTS = {
    # Older style names before smart-split then test with smart
    "file.tar.20240126094501.gz": ("file.tar.20240126094501", ".gz"),
    "file.20240126094501.tar.gz": ("file.20240126094501", ".tar.gz"),
    "file.tar.20240126094501.tar.gz": ("file.tar.20240126094501", ".tar.gz"),
    "file.20240126094501": ("file", ".20240126094501"),
    # This example *shouldn't* happen in general but the following
    # is the correct behavior
    "file.txt.20240126094501": ("file", ".txt.20240126094501"),
    "file.txt.txt.20240126094501": ("file", ".txt.txt.20240126094501"),
    # No extension
    "file.20240126094501": ("file", ".20240126094501"),
}
REG_TEST = {
    # Regular
    "file.jpg": ("file", ".jpg"),
    "file": ("file", ""),
    "file.file": ("file", ".file"),
    # Hidden
    ".file.jpg": (".file", ".jpg"),
    ".file.file": (".file", ".file"),
    ".file": (".file", ""),
    # multiple. Incl. hidden and cap
    "file.tar.gz": ("file", ".tar.gz"),
    "file.TAR.gz": ("file", ".TAR.gz"),
    "file.TAR.GZ": ("file", ".TAR.GZ"),
    ".file.tar.gz": (".file", ".tar.gz"),
    # non-valid second. w/ and w/o valid earlier
    "file.tar.blaaa.gz": ("file.tar.blaaa", ".gz"),  # Stop earlier
    ".file.tar.blaaa.gz": (".file.tar.blaaa", ".gz"),
    "file.tar.blaaa.tar.gz": ("file.tar.blaaa", ".tar.gz"),
    "file.tar.tar.gz": ("file", ".tar.tar.gz"),
    # All valid but keep leading
    ".txt": (".txt", ""),
    ".tar.txt": (".tar", ".txt"),
    ".jpg.gif.tar.txt": (".jpg", ".gif.tar.txt"),
    ".txt.txt": (".txt", ".txt"),
    # Longer path
    "this/has/a.dot/file.ext": ("this/has/a.dot/file", ".ext"),
    "this/has/a.dot/file.jpg.ext": ("this/has/a.dot/file", ".jpg.ext"),
    "this/has/a.dot/noext": ("this/has/a.dot/noext", ""),
}


def test_smart_splitext():
    for inval, goldval in (REG_TEST | DATED_SPLIT_TESTS).items():
        outval = stem, ext = smart_splitext(inval)
        assert goldval == outval, f"Expected {goldval}. Got {outval}"
        assert stem + ext == inval


def test_rpath2apath():
    """
    These also  conditions that can only happen when a user modifies or writes files
    without dfb
    """
    r2a = rpath2apath
    # Normal before smart split
    assert r2a("file.tar.20240126094501.gz") == ("file.tar.gz", 1706262301, "")
    assert r2a("file.tar.20240126094501D.gz") == ("file.tar.gz", 1706262301, "D")
    assert r2a("file.tar.20240126094501R.gz") == ("file.tar.gz", 1706262301, "R")

    # Poorly formed (before smart split)
    rpath2apath("file.tar.gz.20240126094501")

    # Check all of the others
    for filename in DATED_SPLIT_TESTS:  # don't need the split
        apathG = filename.replace(".20240126094501", "")
        apath, ts, tag = r2a(filename)
        assert apath == apathG
        assert ts == 1706262301
        assert tag == ""

        apath, ts, tag = r2a(filename.replace("20240126094501", "20240126094501D"))
        assert apath == apathG
        assert ts == 1706262301
        assert tag == "D"

        apath, ts, tag = r2a(filename.replace("20240126094501", "20240126094501R"))
        assert apath == apathG
        assert ts == 1706262301
        assert tag == "R"

    # Old vs new. These are implied in the others but I want to call it out
    assert (
        r2a("file.tar.20240126094501.gz")
        == r2a("file.20240126094501.tar.gz")
        == r2a("file.tar.gz.20240126094501")
    )
    assert (
        r2a("file.TaR.20240126094501.gz")
        == r2a("file.20240126094501.TaR.gz")
        == r2a("file.TaR.gz.20240126094501")
    )

    assert (
        r2a("file.TaR.20240126094501.gZ")
        == r2a("file.20240126094501.TaR.gZ")
        == r2a("file.TaR.gZ.20240126094501")
    )

    # Test with multiple dates.
    # This is a tough edge case. If the file has a valid datetag already but does *not*
    # have an extension, this should put the tag BEFORE the existing one

    assert r2a("file.19700101000001") == ("file", 1, "")
    assert r2a("file.19700101000002.19700101000001") == ("file.19700101000001", 2, "")
    assert r2a("file.19700101000003.19700101000002.19700101000001") == (
        "file.19700101000003.19700101000001",
        2,
        "",
    )

    # There is still an edge case around something like: "file.19700101000002.jpg.19700101000001"
    # This is ambiguous! The smart-split will call 'jpg.19700101000001' the ext
    # because jpg is a MIME type but it isn't really clear. I am leaving this note but I
    # am also okay with this edge case because it comes from manual ones only!

    assert r2a("file.19700101000002.jpg.19700101000001") == (
        "file.jpg.19700101000001",
        2,
        "",
    )
    # how you get there
    assert (
        apath2rpath("file.jpg.19700101000001", "19700101000002")
        == "file.19700101000002.jpg.19700101000001"
    )


def test_apath2rpath():
    a2r = apath2rpath
    for apath, split in REG_TEST.items():
        gold = split[0] + ".20240126094501" + split[1]
        assert a2r(apath, ts=1706262301, flag="") == gold
        assert a2r(apath, ts=1706262301, flag="D") == gold.replace("501", "501D")
        assert a2r(apath, ts=1706262301, flag="R") == gold.replace("501", "501R")

    apath = "03/gallery4.png@resize=376,812&ssl=1"
    rpath = apath2rpath(apath, ts=1)
    assert rpath2apath(rpath) == (apath, 1, "")

    # Multiple dates. Also test the verify
    assert "file.19700101000001.19700101000002" == apath2rpath(
        "file.19700101000002", "19700101000001"
    )
    assert "file.19700101000001.19700101000002" == apath2rpath(
        "file.19700101000002", "19700101000001", verify=True
    )
    assert "file.19700101000001.19700101000002" == apath2rpath(
        "file.19700101000002", "19700101000001", verify=False
    )


def test_time2all():
    # Make sure it will take any kind of input and give me the same
    ts = int(time.time())
    res = time2all(ts)
    for r in res:
        assert time2all(r).ts == ts


def test_head_tail_table():
    table = ["head", *range(15)]

    # All items
    assert head_tail_table(table, head=0, tail=0, header=True) == table
    assert head_tail_table(table, head=0, tail=0, header=False) == table

    # First head items
    assert head_tail_table(table, head=3, tail=0, header=True) == ["head", 0, 1, 2]
    # Notice this includes the first row since it is saying there is no header
    assert head_tail_table(table, head=3, tail=0, header=False) == ["head", 0, 1]

    # Last tail items
    assert head_tail_table(table, head=0, tail=3, header=True) == ["head", 12, 13, 14]
    head_tail_table(table, head=0, tail=3, header=False) == [12, 13, 14]

    # Too long tail
    assert head_tail_table(table, head=0, tail=15, header=False) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
    ]
    assert head_tail_table(table, head=0, tail=15, header=True) == table
    assert head_tail_table(table, head=0, tail=16, header=True) == table
    assert head_tail_table(table, head=0, tail=16, header=False) == table
    assert head_tail_table(table, head=0, tail=100, header=True) == table
    assert head_tail_table(table, head=0, tail=100, header=False) == table

    # too long head
    assert head_tail_table(table, head=15, tail=0, header=False) == [
        "head",
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
    ]
    assert head_tail_table(table, head=15, tail=0, header=True) == table
    assert head_tail_table(table, head=16, tail=0, header=False) == table
    assert head_tail_table(table, head=16, tail=0, header=True) == table
    assert head_tail_table(table, head=100, tail=0, header=True) == table
    assert head_tail_table(table, head=100, tail=0, header=False) == table

    # head and tail
    head_tail_table(table, head=2, tail=3, header=True) == ["head", 0, 1, 12, 13, 14]
    head_tail_table(table, head=2, tail=3, header=False) == ["head", 0, 12, 13, 14]

    # Overlap
    assert head_tail_table(table, head=8, tail=7, header=True) == table
    assert head_tail_table(table, head=9, tail=7, header=True) == table
    assert head_tail_table(table, head=8, tail=8, header=True) == table
    assert head_tail_table(table, head=9, tail=8, header=True) == table

    # Dots
    assert head_tail_table(table, head=2, tail=3, header=True, dots=True) == [
        "head",
        0,
        1,
        "...",
        12,
        13,
        14,
    ]
    assert head_tail_table(table, head=8, tail=7, header=True, dots=True) == table
    assert head_tail_table(table, head=13, tail=2, header=True, dots=True) == table
    assert head_tail_table(table, head=12, tail=2, header=True, dots=True) == [
        "head",
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        "...",
        13,
        14,
    ]
    assert head_tail_table(table, head=12, tail=1, header=True, dots=True) == [
        "head",
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        "...",
        14,
    ]
    assert head_tail_table(table, head=11, tail=2, header=True, dots=True) == [
        "head",
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        "...",
        13,
        14,
    ]
    assert head_tail_table(table, head=11, tail=4, header=True, dots=True) == table
    assert head_tail_table(table, head=11, tail=4, header=False, dots=True) == [
        "head",
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        "...",
        11,
        12,
        13,
        14,
    ]
    assert head_tail_table(table, head=1, tail=13, dots=True) == [
        "head",
        0,
        "...",
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
    ]
    assert head_tail_table(table, head=1, tail=13, header=False, dots=True) == [
        "head",
        "...",
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
    ]

    # Test with full table. Mostly affects dots
    assert head_tail_table([[1, 2], [3, 4], [5, 6], [7, 8]], header=False, head=1) == [
        [1, 2]
    ]
    assert head_tail_table([[1, 2], [3, 4], [5, 6], [7, 8]], header=False, tail=1) == [
        [7, 8]
    ]
    assert head_tail_table(
        [[1, 2], [3, 4], [5, 6], [7, 8]], header=False, head=1, tail=1
    ) == [[1, 2], [7, 8]]
    assert head_tail_table(
        [[1, 2], [3, 4], [5, 6], [7, 8]], header=False, head=1, tail=1, dots=True
    ) == [[1, 2], ["...", "..."], [7, 8]]


def test_parse_bytes():
    tests = [
        ("23234", 23234),
        ("23234b", 23234),
        ("123k", 123 * 1000),
        ("123 kilobytes", 123 * 1000),
        ("123 kb", 123 * 1000),
        ("123 kib", 123 * 1024),
        ("123 kibi", 123 * 1024),
        ("123 kibibytes", 123 * 1024),
        ("123mi", 128974848),
        ("12.3tB", 12300000000000),
        ("12.3t", 12300000000000),
        ("12.3ti", 13523993021644),
        ("12.3tib", 13523993021644),
        ("0.0002t", 200000000),
        (12345, 12345),
        (12.345, 12),
    ]

    for inval, gold in tests:
        assert parse_bytes(inval) == gold


if __name__ == "__main__":
    # Names and split
    test_smart_splitext()
    test_rpath2apath()
    test_apath2rpath()

    # Others
    test_time2all()
    test_head_tail_table()
    test_parse_bytes()

    print("=" * 50)
    print(" All Passed ".center(50, "="))
    print("=" * 50)



"""
Prune
"""

import datetime
import re
import sys
import bisect
import subprocess
import shlex
import json
import logging
import gzip as gz
from operator import itemgetter

from . import LOCK
from .utils import human_readable_bytes, smart_open
from .timestamps import timestamp_parser
from .dstdb import DFBDST
from .rclonerc import rcpathjoin
from .threadmapper import thread_map_unordered as tmap


logger = logging.getLogger(__name__)


class Prune:
    def __init__(self, config):
        self.config = config
        self.args = config.cliconfig

        if config.disable_prune and not (self.args.dry_run or self.args.dump):
            logger.info("Setting --dry-run based on 'disable_prune = True' ")
            logger.info("Run with `--override 'disable_prune = False'` to override")
            self.args.dry_run = True

        self.dstdb = PruneableDFBDST(config)

    def bydate(self):
        config = self.config
        cliconfig = self.args

        self.when = when = timestamp_parser(
            self.args.when,
            epoch=True,
            now=self.config.now.obj,
        )

        msg = f"Pruning to {timestamp_parser(self.when,aware=True).isoformat()} "
        if self.args.N > 0:
            msg += f"and keeping {self.args.N} additional older version{'s' if self.args.N>1 else ''}"
        if self.args.N < 0:
            msg += f"plus removing {-self.args.N} additional newer version{'s' if self.args.N<-1 else ''}"
        logger.info(msg.strip() + ".")

        self.rpaths = self.dstdb.prune_rpaths(
            self.when,
            self.args.N,
            subdir=self.args.subdir,
        )  # (rpath,size) pairs
        self.rpaths = sorted(self.rpaths)
        if not self.rpaths:
            logger.info("Nothing to prune")
            return

        self.apply()

    def byrpaths(self):
        config = self.config
        cliconfig = self.args

        db = self.dstdb.db()

        err = False
        rpaths = self.rpaths = set()
        for rpath in cliconfig.rpath:
            res = db.execute(
                """
                SELECT rpath,size,ref_rpath,isref
                FROM items
                WHERE 
                    rpath = ?
                """,
                (rpath,),
            ).fetchall()

            if not res:
                logger.warning(f"No matches for {rpath!r}")
                continue

            reg, ref = [], []
            for row in res:
                {True: ref, False: reg}[bool(row["isref"])].append(row)

            if len(reg) > 1:
                logger.debug(f"Multiple non-references to {rpath = }")

            msg = f"{rpath!r} has {len(res)} entr{'ies' if len(res)>1 else 'y'}. "
            msg += "{len(reg)} regular and {len(ref)} references"
            logger.debug(msg)

            for row in reg:
                rpaths.add((row["rpath"], row["size"]))
                logger.debug(" delete {row['rpath']!r}")

            for row in ref:
                msg = f"Deleteing {rpath!r} will break {row['ref_rpath']!r}."
                if cliconfig.error_if_referenced:
                    err = True
                    logger.error(msg + " Will not continue")
                else:
                    # Add them. Not really an "rpath" but will have the same effect of
                    # deleting the file.
                    rpaths.add((row["ref_rpath"], 0))
                    logger.debug(" delete {row['ref_rpath']!r}")
        if err:
            msg = "References will break. Use '--no-error-if-referenced' flag to force."
            raise BrokenReferenceError(msg)

        self.apply()

    def apply(self):
        self.summary()
        config = self.config
        cliconfig = self.args

        rpaths = (r[0] for r in self.rpaths)

        if file := cliconfig.dump:
            logger.info(f"Dumping jsonl then exiting")
            try:
                fp = smart_open(file, "wt") if file != "-" else sys.stdout
                for rpath in rpaths:
                    item = {"_V": 1, "_action": "prune", "rpath": rpath}
                    print(
                        json.dumps(item, ensure_ascii=False, separators=(",", ":")),
                        file=fp,
                        flush=True,
                    )
            finally:
                if file != "-":
                    fp.close()
                    logger.info(f"Written to {file!r}")
            return

        if self.args.dry_run:
            logger.info("DRY-RUN. Exit")
            return
        elif self.args.interactive:
            r = input("Do you want to continue? [Y]/N:")
            if r.lower().startswith("n"):
                return

        self.errcount = 0

        rc = self.config.rc
        rc.start()

        def _delete(rpath):
            try:
                logger.info(f"Pruning {rpath!r}.")
                rc.delete((self.config.dst, rpath))
                return rpath
            except subprocess.CalledProcessError as EE:
                logger.error(f"Could not prune {rpath!r}. {EE}")
                with LOCK:
                    self.errcount += 1

        rpaths = tmap(_delete, rpaths, Nt=self.config.concurrency)
        rpaths = filter(bool, rpaths)  # Remove errors
        rpaths = map(self.dstdb.delete_rpath, rpaths)  # on main thread only
        for _ in rpaths:
            pass

        self.dstdb.push_snapshots()

    def summary(self):
        _p = logger.debug
        if self.args.dry_run or self.args.interactive:
            _p = logger.info

        num, units = human_readable_bytes(sum(r[-1] for r in self.rpaths if r[-1] >= 0))
        s = "s" if len(self.rpaths) != 1 else ""
        logger.info(f"Pruning {len(self.rpaths)} file{s} ({num:0.2f} {units})")

        for rpath, size in self.rpaths:
            num, units = human_readable_bytes(size)
            paren = f"{num:0.2f} {units}" if size >= 0 else "DEL"
            _p(f"    {rpath!r} ({paren})")


class PruneableDFBDST(DFBDST):
    def prune_rpaths(self, when, keep, subdir=""):
        # Pruning is more complex than it seems at first because of reference files.
        # We do not want to delete files still being referenced. We also don't want
        # to delete a delete-marker that "hides" those still-referenced files.
        # The algorithm to do this is described in comments
        #
        # When subdir is selected, we can consider files that are in the subdir OR
        # point to it (for references). Note, previous behavior was WRONG and would
        # deleted referent apaths when a reference was outside subdir. This is fixed
        # (and tested)
        subdir = f"{subdir.removesuffix('/').removeprefix('./')}/".removeprefix("/")

        # Step 0: Group by aname. These are sorted by timestamp. Only look within the
        #         subdir but will later add things that point inside.
        groups = self.group_by_apath(
            select="rpath,apath,timestamp,size,ref_rpath",
            conditions=("WHERE apath LIKE ?", [subdir + "%"]),
        )

        ## Regular prune -- This would be it if not for references
        # Step 1a: Bisect the group to find the cutoff spot, when <= ts
        #          (<= means bisect_right). Then do icut = iwhen - 1
        #          to account for keeping the element older.
        # Step 1b: Shift by "keep" to keep specified versions
        # Step 1c: If there is nothing before the cutoff, keep it all (icut = 0) and
        #          make sure to also keep at least one unless that one is delete.
        #          If it is delete, it will be covered in step 2
        # Step 1d: Keep ~~everything~~ (NO) to the right of icut. Universal set
        #          Only keep references to the right.
        # Step 1e: Save for deletetion group to the left of icut. Grouped list.
        #          Make sure to include ref_rpaths too. They will always get deleted.
        # Step 1f: If subdir, add in additional references to keep that could point
        #          there from outside only
        keep_rpaths = set()
        del_groups = {}
        for name, group in groups:
            iwhen = keyed_bisect_right(group, when, key=itemgetter("timestamp"))  # 1a
            iwhen -= keep  # 1b
            if iwhen >= len(group) and group[-1]["size"] < 0:
                icut = iwhen
            else:
                icut = max([iwhen - 1, 0])  # 1a,c
                icut = min([icut, len(group) - 1])

            keep_rpaths.update(
                row["rpath"] for row in group[icut:] if row.get("ref_rpath", None)
            )  # 1d
            logger.debug(f'(1) {name!r} keep {[row["rpath"] for row in group[icut:]]}')

            del_groups[name] = group[:icut]  # 1e
            r0 = [row["rpath"] for row in group[:icut]]
            r1 = [row.get("ref_rpath", None) for row in group[:icut]]
            logger.debug(f"(1) {name!r} consider for del {r0 + list(filter(bool,r1))}")

        if subdir:  # 1f
            with self.db() as db:
                res = db.execute(
                    """
                    SELECT rpath
                    FROM items
                    WHERE
                        rpath LIKE ? -- points inside subdir
                    AND
                        apath NOT LIKE ? -- but ins't inside it
                    """,
                    (subdir + "%", subdir + "%"),
                )
                keep_rpaths.update(row["rpath"] for row in res)

        # Step 2:  Loop over each group of to-be-deleted files
        # Step 2a: Delete everything that isn't (1) referenced (i.e. in keep_rpaths)
        #          or (2) a delete marker. Includes ref_rpath; which is not rpath
        #          but will delete the reference file and do nothing when applied to the
        #          DB (but the actual rpath will have been deleted too)
        # Step 2b: Remove any delete markers if and only if not the last item since
        #          they aren't hiding anything except maybe the final.
        # Step 2c: If and only if there is only ONE remaining item and it is a delete
        #          marker, delete it. (This makes sure that a delete marker "hides"
        #          a kept file, that is kept because of reference)
        del_rpaths = set()
        for name, group in del_groups.items():  # only with subdir from step 0
            _d = set()  # This isn't needed but makes debug msg easier
            keep_group = []
            for row in group:  # 2a
                if rr := row.get("ref_rpath", None):
                    _d.add((rr, 0))

                if row["rpath"] in keep_rpaths or row["size"] < 0:  # (1) & (2)
                    keep_group.append(row)  # do not delete yet
                    continue

                _d.add((row["rpath"], row["size"]))

            del_rpaths.update(_d)

            logger.debug(
                f'(2a) {name!r} temp keep {[row["rpath"] for row in keep_group]}'
            )
            logger.debug(f"(2a) {name!r} del {_d}")

            if not keep_group:
                continue

            # 2b
            _d = set()
            still_keep = []
            for row in keep_group[:-1]:  # 2b. Do not consider the last row
                if row["size"] < 0:
                    _d.add((row["rpath"], row["size"]))
                else:
                    still_keep.append(row)
            still_keep.append(keep_group[-1])  # Add the last item back
            del_rpaths.update(_d)

            logger.debug(
                f'(2b) {name!r} temp keep {[row["rpath"] for row in still_keep]}'
            )
            logger.debug(f"(2b) {name!r} del {_d}")

            # 2c
            if len(still_keep) > 1:
                continue
            row = still_keep[0]
            if row["size"] < 0:
                del_rpaths.add((row["rpath"], row["size"]))

                logger.debug(f'(2c) {name!r} del {row["rpath"]!r}')

        # A note: This can be made even more agressive because this may leave behind an
        #         unneeded delete marker. For the most part, this is ignore but see
        #         if there are any

        # # Not used yet. Needs to be more thought out and tested
        # res = self.db().execute("""
        #     SELECT rpath
        #     FROM (
        #         -- Nest the conditional so that it finds all initials and then
        #         -- later takes deletes.
        #         SELECT rpath,size
        #         FROM items
        #         GROUP BY apath
        #         HAVING MIN(timestamp) -- NOT max
        #     )
        #     WHERE size <0""")
        # for row in res:
        #     del_rpaths.add((row["rpath"], -1))

        # This section has 100% test coverage! Very good to have with the complexities
        return del_rpaths

    def delete_rpath(self, rpath):
        """
        Remove all entries that point to the rpath even if they are references
        """
        db = self.db()
        with db:
            db.execute(
                """
                DELETE FROM items 
                WHERE rpath = ?""",
                (rpath,),
            )
        db.commit()
        db.close()

        item = {"_V": 1, "_action": "prune", "rpath": rpath}

        with self.snap_file.open(mode="at") as fp:
            print(json.dumps(item), file=fp, flush=True)

        return rpath


class BrokenReferenceError(ValueError):
    pass


class KeyedListWrapper:
    """
    Create a wrapper for a list with keys. This isn't needed in Python 3.10+
    Based on https://gist.github.com/ericremoreynolds/2d80300dabc70eebc790
    """

    def __init__(self, mylist, key, cache=False):
        self.mylist = mylist
        self.key = key
        if cache:
            self.cache = [self.key(item) for item in mylist]
        else:
            self.cache = []

    def __len__(self):
        return len(self.mylist)

    def __getitem__(self, ix):
        try:
            return self.cache[ix]
        except IndexError:
            return self.key(self.mylist[ix]) if self.key else self.mylist[ix]


def keyed_bisect(a, x, *, key=None, cache=False, **kwargs):
    return bisect.bisect(KeyedListWrapper(a, key, cache=cache), x, **kwargs)


keyed_bisect_right = keyed_bisect


def keyed_bisect_left(a, x, *, key=None, cache=False, **kwargs):
    return bisect.bisect_left(KeyedListWrapper(a, key, cache=cache), x, **kwargs)
